"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import HypeTrainEventsResponseContributions, HypeTrainEventsResponseData

__all__ = ("HypeTrainContribution", "HypeTrainEvent")


class HypeTrainEvent:
    """Represents a Hype Train Event.

    Attributes
    -----------
    id: str
        The ID of the event.
    hype_train_id: str
        The ID of the Hype Train.
    type: str
        The type of the event, in the form, hypetrain.{event_name} (i.e. `hypetrain.progression`).
    version: str
        The version of the endpoint.
    broadcaster: PartialUser
        The user whose channel the Hype Train is occurring on.
    timestamp: datetime.datetime
        The time the event happened at.
    cooldown_end_time: datetime.datetime
        The time that another Hype Train can happen at.
    expiry: datetime.datetime
        The time that this Hype Train expires at.
    started_at: datetime.datetime
        The time that this Hype Train started at.
    last_contribution: HypeTrainContribution
        The last contribution to this Hype Train.
    level: int
        The level reached on this Hype Train (1-5).
    top_contributions: list[HypeTrainContribution]
        The top contributors to the Hype Train.
    total_contributed: int
        The current total amount raised.
    goal: int
        The goal for the next Hype Train level
    """

    __slots__ = (
        "broadcaster",
        "cooldown_end_time",
        "expiry",
        "goal",
        "hype_train_id",
        "id",
        "last_contribution",
        "level",
        "started_at",
        "timestamp",
        "top_contributions",
        "total_contributed",
        "type",
        "version",
    )

    def __init__(self, data: HypeTrainEventsResponseData, *, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.type: str = data["event_type"]
        self.timestamp: datetime.datetime = parse_timestamp(data["event_timestamp"])
        self.version: str = data["version"]
        self.broadcaster = PartialUser(data["event_data"]["broadcaster_id"], None, http=http)
        self.cooldown_end_time: datetime.datetime = parse_timestamp(data["event_data"]["cooldown_end_time"])
        self.expiry: datetime.datetime = parse_timestamp(data["event_data"]["expires_at"])
        self.goal: int = int(data["event_data"]["goal"])
        self.hype_train_id: str = data["event_data"]["id"]
        self.last_contribution = HypeTrainContribution(data["event_data"]["last_contribution"], http=http)
        self.level: int = int(data["event_data"]["level"])
        self.started_at: datetime.datetime = parse_timestamp(data["event_data"]["started_at"])
        self.top_contributions = [HypeTrainContribution(d, http=http) for d in data["event_data"]["top_contributions"]]
        self.total_contributed: int = data["event_data"]["total"]

    def __repr__(self) -> str:
        return f"<HypeTrainEvent id={self.id} type={self.type} level={self.level} broadcaster={self.broadcaster}>"


class HypeTrainContribution:
    """A Contribution to a Hype Train

    Attributes
    -----------
    total: int
        The total amount contributed. If type is ``BITS``, total represents the amount of Bits used.
        If type is ``SUBS``, total is 500, 1000, or 2500 to represent tier 1, 2, or 3 subscriptions, respectively.
    type: typing.Literal["BITS", "SUBS", "OTHER"]
        Identifies the contribution method, either BITS, SUBS or OTHER.
    user: PartialUser
        The user making the contribution.
    """

    __slots__ = "total", "type", "user"

    def __init__(self, data: HypeTrainEventsResponseContributions, *, http: HTTPClient) -> None:
        self.total: int = int(data["total"])
        self.type: Literal["BITS", "SUBS", "OTHER"] = data["type"]
        self.user = PartialUser(data["user"], None, http=http)

    def __repr__(self) -> str:
        return f"<HypeTrainContribution total={self.total} type={self.type} user={self.user}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import abc
import asyncio
import datetime
import enum
from collections.abc import Callable, Coroutine, Hashable
from typing import TYPE_CHECKING, Any, Generic, Self, TypeAlias, TypeVar


if TYPE_CHECKING:
    import twitchio

    from .context import Context


__all__ = ("BaseCooldown", "Bucket", "BucketType", "Cooldown", "GCRACooldown")


PT = TypeVar("PT")
CT = TypeVar("CT")


class BucketType(enum.Enum):
    """Enum representing default implementations for the key argument in :func:`~.commands.cooldown`.

    Attributes
    ----------
    default
        The cooldown will be considered a global cooldown shared across every channel and user.
    user
        The cooldown will apply per user, accross all channels.
    channel
        The cooldown will apply to every user/chatter in the channel.
    chatter
        The cooldown will apply per user, per channel.
    """

    default = 0
    user = 1
    channel = 2
    chatter = 3

    def get_key(self, payload: twitchio.ChatMessage | Context) -> Any:
        if self is BucketType.user:
            return payload.chatter.id

        elif self is BucketType.channel:
            return ("channel", payload.broadcaster.id)

        elif self is BucketType.chatter:
            return (payload.broadcaster.id, payload.chatter.id)

    def __call__(self, payload: twitchio.ChatMessage | Context) -> Any:
        return self.get_key(payload)


class BaseCooldown(abc.ABC):
    """Base class used to implement your own cooldown algorithm for use with :func:`~.commands.cooldown`.

    Some built-in cooldown algorithms already exist:

    - :class:`~.commands.Cooldown` - (``Token Bucket Algorithm``)

    - :class:`~.commands.GCRACooldown` - (``Generic Cell Rate Algorithm``)


    .. note::

        Every base method must be implemented in this base class.
    """

    @abc.abstractmethod
    def reset(self) -> None:
        """Base method which should be implemented to reset the cooldown."""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, *args: Any, **kwargs: Any) -> float | None:
        """Base method which should be implemented to update the cooldown/ratelimit.

        This is where your algorithm logic should be contained.

        .. important::

            This method should always return a :class:`float` or ``None``. If ``None`` is returned by this method,
            the cooldown will be considered bypassed.

        Returns
        -------
        :class:`float`
            The time needed to wait before you are off cooldown.
        ``None``
            Bypasses the cooldown.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self) -> Self:
        """Base method which should be implemented to return a copy of this class in it's original state."""
        raise NotImplementedError

    @abc.abstractmethod
    def is_ratelimited(self, *args: Any, **kwargs: Any) -> bool:
        """Base method which should be implemented which returns a bool indicating whether the cooldown is ratelimited.

        Returns
        -------
        bool
            A bool indicating whether this cooldown is currently ratelimited.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_dead(self, *args: Any, **kwargs: Any) -> bool:
        """Base method which should be implemented to indicate whether the cooldown should be considered stale and allowed
        to be removed from the ``bucket: cooldown`` mapping.

        Returns
        -------
        bool
            A bool indicating whether this cooldown is stale/old.
        """
        raise NotImplementedError


class Cooldown(BaseCooldown):
    """Default cooldown algorithm for :func:`~.commands.cooldown`, which implements a ``Token Bucket Algorithm``.

    See: :func:`~.commands.cooldown` for more documentation.
    """

    def __init__(self, *, rate: int, per: float | datetime.timedelta) -> None:
        if rate <= 0:
            raise ValueError(f'Cooldown rate must be equal to or greater than 1. Got "{rate}" expected >= 1.')

        self._rate: int = rate
        self._per: datetime.timedelta = datetime.timedelta(seconds=per) if not isinstance(per, datetime.timedelta) else per

        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        self._window: datetime.datetime = now + self._per

        if self._window <= now:
            raise ValueError("The provided per value for Cooldowns can not go into the past.")

        self._tokens: int = self._rate
        self.last_updated: datetime.datetime | None = None

    @property
    def per(self) -> datetime.timedelta:
        return self._per

    def reset(self) -> None:
        self._tokens = self._rate
        self._window = datetime.datetime.now(tz=datetime.UTC) + self._per

    def get_tokens(self, now: datetime.datetime | None = None) -> int:
        if now is None:
            now = datetime.datetime.now(tz=datetime.UTC)

        tokens = max(self._tokens, 0)
        if now > self._window:
            tokens = self._rate

        return tokens

    def is_ratelimited(self) -> bool:
        self._tokens = self.get_tokens()
        return self._tokens == 0

    def update(self, *, factor: int = 1) -> float | None:
        now = datetime.datetime.now(tz=datetime.UTC)
        self.last_updated = now

        self._tokens = self.get_tokens(now)

        if self._tokens == self._rate:
            self._window = datetime.datetime.now(tz=datetime.UTC) + self._per

        self._tokens -= factor

        if self._tokens < 0:
            remaining = (self._window - now).total_seconds()
            return remaining

    def copy(self) -> Self:
        return self.__class__(rate=self._rate, per=self._per)

    def is_dead(self) -> bool:
        if self.last_updated is None:
            return False

        now = datetime.datetime.now(tz=datetime.UTC)
        return now > (self.last_updated + self.per)


class GCRACooldown(BaseCooldown):
    """GCRA cooldown algorithm for :func:`~.commands.cooldown`, which implements the ``GCRA`` ratelimiting algorithm.

    See: :func:`~.commands.cooldown` for more documentation.
    """

    def __init__(self, *, rate: int, per: float | datetime.timedelta) -> None:
        if rate <= 0:
            raise ValueError(f'Cooldown rate must be equal to or greater than 1. Got "{rate}" expected >= 1.')

        self._rate: int = rate
        self._per: datetime.timedelta = datetime.timedelta(seconds=per) if not isinstance(per, datetime.timedelta) else per

        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        self._tat: datetime.datetime | None = None

        if (now + self._per) <= now:
            raise ValueError("The provided per value for Cooldowns can not go into the past.")

        self.last_updated: datetime.datetime | None = None

    @property
    def inverse(self) -> float:
        return self._per.total_seconds() / self._rate

    @property
    def per(self) -> datetime.timedelta:
        return self._per

    def reset(self) -> None:
        self.last_updated = None
        self._tat = None

    def is_ratelimited(self, *, now: datetime.datetime | None = None) -> bool:
        now = now or datetime.datetime.now(tz=datetime.UTC)
        tat: datetime.datetime = max(self._tat or now, now)

        separation: float = (tat - now).total_seconds()
        max_interval: float = self._per.total_seconds() - self.inverse

        return separation > max_interval

    def update(self) -> float | None:
        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        tat: datetime.datetime = max(self._tat or now, now)

        self.last_updated = now

        separation: float = (tat - now).total_seconds()
        max_interval: float = self._per.total_seconds() - self.inverse

        if separation > max_interval:
            return separation - max_interval

        new = max(tat, now) + datetime.timedelta(seconds=self.inverse)
        self._tat = new

    def copy(self) -> Self:
        return self.__class__(rate=self._rate, per=self._per)

    def is_dead(self) -> bool:
        if self.last_updated is None:
            return False

        now = datetime.datetime.now(tz=datetime.UTC)
        return now > (self.last_updated + self.per)


KeyT: TypeAlias = Callable[..., Hashable] | Callable[..., Coroutine[Any, Any, Hashable]] | BucketType


class Bucket(Generic[PT]):
    def __init__(self, cooldown: BaseCooldown, *, key: KeyT) -> None:
        self._cooldown: BaseCooldown = cooldown
        self._cache: dict[Hashable, BaseCooldown] = {}
        self._key: KeyT = key

    @classmethod
    def from_cooldown(cls, *, base: type[BaseCooldown], key: KeyT, **kwargs: Any) -> Self:
        cd: BaseCooldown = base(**kwargs)
        return cls(cd, key=key)

    def create_cooldown(self) -> BaseCooldown | None:
        return self._cooldown.copy()

    def verify_cache(self) -> None:
        dead = [k for k, v in self._cache.items() if v.is_dead()]
        for key in dead:
            del self._cache[key]

    async def get_key(self, payload: PT) -> Hashable:
        if asyncio.iscoroutinefunction(self._key):
            key = await self._key(payload)  # type: ignore
        else:
            key = self._key(payload)  # type: ignore

        return key

    async def get_cooldown(self, payload: PT) -> BaseCooldown | None:
        if self._key is BucketType.default:
            return self._cooldown

        self.verify_cache()
        key = await self.get_key(payload)
        if key is None:
            return

        if key not in self._cache:
            cooldown = self.create_cooldown()

            if cooldown is not None:
                self._cache[key] = cooldown
        else:
            cooldown = self._cache[key]

        return cooldown

    async def update(self, payload: PT, **kwargs: Any) -> float | None:
        bucket = await self.get_cooldown(payload)

        if bucket is None:
            return None

        return bucket.update(**kwargs)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import twitchio

    from .authentication import UserTokenPayload
    from .models.eventsub_ import SubscriptionRevoked
    from .payloads import TokenRefreshedPayload

async def event_token_refreshed(payload: TokenRefreshedPayload) -> None:
    """Event dispatched when a token managed by the :class:`~twitchio.Client` is successfully refreshed.

    You can use this event to update the stored token locally.

    Parameters
    ----------
    payload: TokenRefreshedPayload
    """

async def event_oauth_authorized(payload: UserTokenPayload) -> None:
    """Event dispatched when a user authorizes your Client-ID via Twitch OAuth on a built-in web adapter.

    The default behaviour of this event is to add the authorized token to the client.
    See: :class:`~twitchio.Client.add_token` for more details.

    Parameters
    ----------
    payload: UserTokenPayload
    """

async def event_subscription_revoked(payload: SubscriptionRevoked) -> None:
    """Event dispatched when Twitch revokes a subscription. This can occur on websockets and webhooks.
    You'll receive this message once and then no longer receive messages for the specified user and subscription type.

    Webhook:
        - The user mentioned in the subscription no longer exists. The notification's status is set to `user_removed`.
        - The user revoked the authorization token or simply changed their password. The notification's status is set to `authorization_revoked`.
        - The callback failed to respond in a timely manner too many times. The notification's status is set to `notification_failures_exceeded`.
        - The subscribed to subscription type and version is no longer supported. The notification's status is set to `version_removed`.

    Websocket:
        - The user mentioned in the subscription no longer exists. The notification's status field is set to `user_removed`.
        - The user revoked the authorization token that the subscription relied on. The notification's status field is set to `authorization_revoked`.
        - The subscribed to subscription type and version is no longer supported. The notification's status field is set to `version_removed`.

    Parameters
    ----------
    payload: SubscriptionRevoked
    """

async def event_ready() -> None:
    """Event dispatched when the Client is ready and has completed login."""

# Eventsub Events

# AutoMod
async def event_automod_message_hold(payload: twitchio.AutomodMessageHold) -> None: ...
async def event_automod_message_update(payload: twitchio.AutomodMessageUpdate) -> None: ...
async def event_automod_settings_update(payload: twitchio.AutomodSettingsUpdate) -> None: ...
async def event_automod_terms_update(payload: twitchio.AutomodTermsUpdate) -> None: ...

# Channel/Broadcaster
async def event_channel_update(payload: twitchio.ChannelUpdate) -> None: ...
async def event_follow(payload: twitchio.ChannelFollow) -> None: ...
async def event_ad_break(payload: twitchio.ChannelAdBreakBegin) -> None: ...
async def event_cheer(payload: twitchio.ChannelCheer) -> None: ...
async def event_raid(payload: twitchio.ChannelRaid) -> None: ...

# Chat/Messages
async def event_message(payload: twitchio.ChatMessage) -> None: ...
async def event_message_whisper(payload: twitchio.Whisper) -> None: ...
async def event_message_delete(payload: twitchio.ChatMessageDelete) -> None: ...
async def event_chat_notification(payload: twitchio.ChatNotification) -> None: ...
async def event_chat_clear(payload: twitchio.ChannelChatClear) -> None: ...
async def event_chat_clear_user(payload: twitchio.ChannelChatClearUserMessages) -> None: ...
async def event_chat_settings_update(payload: twitchio.ChatSettingsUpdate) -> None: ...
async def event_chat_user_message_hold(payload: twitchio.ChatUserMessageHold) -> None: ...
async def event_chat_user_message_update(payload: twitchio.ChatUserMessageUpdate) -> None: ...
async def event_bits_use(payload: twitchio.ChannelBitsUse) -> None: ...

# Shared Chat
async def event_shared_chat_begin(payload: twitchio.SharedChatSessionBegin) -> None: ...
async def event_shared_chat_update(payload: twitchio.SharedChatSessionUpdate) -> None: ...
async def event_shared_chat_end(payload: twitchio.SharedChatSessionEnd) -> None: ...

# Subscriptions
async def event_subscription(payload: twitchio.ChannelSubscribe) -> None: ...
async def event_subscription_end(payload: twitchio.ChannelSubscriptionEnd) -> None: ...
async def event_subscription_gift(payload: twitchio.ChannelSubscriptionGift) -> None: ...
async def event_subscription_message(payload: twitchio.ChannelSubscriptionMessage) -> None: ...

# Bans
async def event_ban(payload: twitchio.ChannelBan) -> None: ...
async def event_unban(payload: twitchio.ChannelUnban) -> None: ...
async def event_unban_request(payload: twitchio.ChannelUnbanRequest) -> None: ...
async def event_unban_request_resolve(payload: twitchio.ChannelUnbanRequestResolve) -> None: ...

# Warnings
async def event_warning_acknowledge(payload: twitchio.ChannelWarningAcknowledge) -> None: ...
async def event_warning_send(payload: twitchio.ChannelWarningSend) -> None: ...

# Moderation and VIPs
async def event_mod_action(payload: twitchio.ChannelModerate) -> None: ...
async def event_moderator_add(payload: twitchio.ChannelModeratorAdd) -> None: ...
async def event_moderator_remove(payload: twitchio.ChannelModeratorRemove) -> None: ...
async def event_vip_add(payload: twitchio.ChannelVIPAdd) -> None: ...
async def event_vip_remove(payload: twitchio.ChannelVIPRemove) -> None: ...

# Redemptions and Rewards
async def event_automatic_redemption_add(payload: twitchio.ChannelPointsAutoRedeemAdd) -> None: ...
async def event_custom_reward_add(payload: twitchio.ChannelPointsRewardAdd) -> None: ...
async def event_custom_reward_update(payload: twitchio.ChannelPointsRewardUpdate) -> None: ...
async def event_custom_reward_remove(payload: twitchio.ChannelPointsRewardRemove) -> None: ...
async def event_custom_redemption_add(payload: twitchio.ChannelPointsRedemptionAdd) -> None: ...
async def event_custom_redemption_update(payload: twitchio.ChannelPointsRedemptionUpdate) -> None: ...

# Polls
async def event_poll_begin(payload: twitchio.ChannelPollBegin) -> None: ...
async def event_poll_progress(payload: twitchio.ChannelPollProgress) -> None: ...
async def event_poll_end(payload: twitchio.ChannelPollEnd) -> None: ...

# Predictions

async def event_prediction_begin(payload: twitchio.ChannelPredictionBegin) -> None: ...
async def event_prediction_progress(payload: twitchio.ChannelPredictionProgress) -> None: ...
async def event_prediction_lock(payload: twitchio.ChannelPredictionLock) -> None: ...
async def event_prediction_end(payload: twitchio.ChannelPredictionEnd) -> None: ...

# Suspicious Users
async def event_suspicious_user_message(payload: twitchio.SuspiciousUserMessage) -> None: ...
async def event_suspicious_user_update(payload: twitchio.SuspiciousUserUpdate) -> None: ...

# Charity Campaigns
async def event_charity_campaign_donate(payload: twitchio.CharityCampaignDonation) -> None: ...
async def event_charity_campaign_start(payload: twitchio.CharityCampaignStart) -> None: ...
async def event_charity_campaign_progress(payload: twitchio.CharityCampaignProgress) -> None: ...
async def event_charity_campaign_stop(payload: twitchio.CharityCampaignStop) -> None: ...

# Goals
async def event_goal_begin(payload: twitchio.GoalBegin) -> None: ...
async def event_goal_progress(payload: twitchio.GoalProgress) -> None: ...
async def event_goal_end(payload: twitchio.GoalEnd) -> None: ...

# Hype Train
async def event_hype_train(payload: twitchio.HypeTrainBegin) -> None: ...
async def event_hype_train_progress(payload: twitchio.HypeTrainProgress) -> None: ...
async def event_hype_train_end(payload: twitchio.HypeTrainEnd) -> None: ...

# Shield Mode
async def event_shield_mode_begin(payload: twitchio.ShieldModeBegin) -> None: ...
async def event_shield_mode_end(payload: twitchio.ShieldModeEnd) -> None: ...

# Shoutouts
async def event_shoutout_create(payload: twitchio.ShoutoutCreate) -> None: ...
async def event_shoutout_receive(payload: twitchio.ShoutoutReceive) -> None: ...

# Streams
async def event_stream_online(payload: twitchio.StreamOnline) -> None: ...
async def event_stream_offline(payload: twitchio.StreamOffline) -> None: ...

# OAuth
async def event_user_authorization_grant(payload: twitchio.UserAuthorizationGrant) -> None: ...
async def event_user_authorization_revoke(payload: twitchio.UserAuthorizationRevoke) -> None: ...

# Users
async def event_user_update(payload: twitchio.UserUpdate) -> None: ...



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from twitchio.assets import Asset


if TYPE_CHECKING:
    from twitchio.http import HTTPClient
    from twitchio.types_.responses import GamesResponseData


__all__ = ("Game",)


class Game:
    """Represents a Game on Twitch.

    You can retrieve a game by its ID, name or IGDB ID using the :meth:`~twitchio.Client.fetch_games`
    method or the various ``.fetch_game()`` methods of other models.

    To fetch a list of games, see: :meth:`~twitchio.Client.fetch_games`

    Supported Operations
    --------------------

    +-------------+----------------------------------+----------------------------------------------------+
    | Operation   | Usage(s)                        | Description                                         |
    +=============+==================================+====================================================+
    | __str__     | str(game), f"{game}"             | Returns the games name.                            |
    +-------------+----------------------------------+----------------------------------------------------+
    | __repr__    | repr(game), f"{game!r}"          | Returns the games official representation.         |
    +-------------+----------------------------------+----------------------------------------------------+
    | __eq__      | game == game2, game != game2     | Checks if two games are equal.                     |
    +-------------+----------------------------------+----------------------------------------------------+


    Attributes
    ----------
    id: str
        The ID of the game provided by Twitch.
    name: str
        The name of the game.
    box_art: Asset
        The box art of the game as an :class:`~twitchio.Assets`.
    igdb_id: str | None
        The IGDB ID of the game. If this is not available to Twitch it will be ``None``.
    """

    __slots__ = ("box_art", "id", "igdb_id", "name")

    def __init__(self, data: GamesResponseData, *, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.igdb_id: str | None = data.get("igdb_id", None)
        self.box_art: Asset = Asset(data["box_art_url"], http=http, dimensions=(1080, 1440))

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<Game id={self.id} name={self.name}>"

    def __eq__(self, __value: object) -> bool:
        return __value.id == self.id if isinstance(__value, Game) else NotImplemented



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import importlib.util
import logging
import sys
import types
from typing import TYPE_CHECKING, Any, TypeAlias, Unpack

from twitchio.client import Client

from ...utils import _is_submodule
from .context import Context
from .converters import _BaseConverter
from .core import Command, CommandErrorPayload, Group, Mixin
from .exceptions import *


if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Iterable, Mapping

    from twitchio.eventsub.subscriptions import SubscriptionPayload
    from twitchio.models.eventsub_ import ChannelPointsRedemptionAdd, ChannelPointsRedemptionUpdate, ChatMessage
    from twitchio.types_.eventsub import SubscriptionResponse
    from twitchio.user import PartialUser

    from .components import Component
    from .types_ import BotOptions

    PrefixT: TypeAlias = str | Iterable[str] | Callable[["Bot", "ChatMessage"], Coroutine[Any, Any, str | Iterable[str]]]


logger: logging.Logger = logging.getLogger(__name__)


class Bot(Mixin[None], Client):
    """The TwitchIO ``commands.Bot`` class.

    The Bot is an extension of and inherits from :class:`twitchio.Client` and comes with additonal powerful features for
    creating and managing bots on Twitch.

    Unlike :class:`twitchio.Client`, the :class:`~.Bot` class allows you to easily make use of built-in the commands ext.

    The easiest way of creating and using a bot is via subclassing, some examples are provided below.

    .. note::

        Any examples contained in this class which use ``twitchio.Client`` can be changed to ``commands.Bot``.


    Parameters
    ----------
    client_id: str
        The client ID of the application you registered on the Twitch Developer Portal.
    client_secret: str
        The client secret of the application you registered on the Twitch Developer Portal.
        This must be associated with the same ``client_id``.
    bot_id: str
        The User ID associated with the Bot Account.
        Unlike on :class:`~twitchio.Client` this is a required argument on :class:`~.Bot`.
    owner_id: str | None
        An optional ``str`` which is the User ID associated with the owner of this bot. This should be set to your own user
        accounts ID, but is not required. Defaults to ``None``.
    prefix: str | Iterabale[str] | Coroutine[Any, Any, str | Iterable[str]]
        The prefix(es) to listen to, to determine whether a message should be treated as a possible command.

        This can be a ``str``, an iterable of ``str`` or a coroutine which returns either.

        This is a required argument, common prefixes include: ``"!"`` or ``"?"``.

    Example
    -------

        .. code:: python3

            import asyncio
            import logging

            import twitchio
            from twitchio import eventsub
            from twitchio.ext import commands

            LOGGER: logging.Logger = logging.getLogger("Bot")

            class Bot(commands.Bot):

                def __init__(self) -> None:
                    super().__init__(client_id="...", client_secret="...", bot_id="...", owner_id="...", prefix="!")

                # Do some async setup, as an example we will load a component and subscribe to some events...
                # Passing the bot to the component is completely optional...
                async def setup_hook(self) -> None:

                    # Listen for messages on our channel...
                    # You need appropriate scopes, see the docs on authenticating for more info...
                    payload = eventsub.ChatMessageSubscription(broadcaster_user_id=self.owner_id, user_id=self.bot_id)
                    await self.subscribe_websocket(payload=payload)

                    await self.add_component(SimpleCommands(self))
                    LOGGER.info("Finished setup hook!")

            class SimpleCommands(commands.Component):

                def __init__(self, bot: Bot) -> None:
                    self.bot = bot

                @commands.command()
                async def hi(self, ctx: commands.Context) -> None:
                    '''Command which sends you a hello.'''
                    await ctx.reply(f"Hello {ctx.chatter}!")

                @commands.command()
                async def say(self, ctx: commands.Context, *, message: str) -> None:
                    '''Command which repeats what you say: !say I am an apple...'''
                    await ctx.send(message)

            def main() -> None:
                # Setup logging, this is optional, however a nice to have...
                twitchio.utils.setup_logging(level=logging.INFO)

                async def runner() -> None:
                    async with Bot() as bot:
                        await bot.start()

                try:
                    asyncio.run(runner())
                except KeyboardInterrupt:
                    LOGGER.warning("Shutting down due to Keyboard Interrupt...")

            main()
    """

    def __init__(
        self,
        *,
        client_id: str,
        client_secret: str,
        bot_id: str,
        owner_id: str | None = None,
        prefix: PrefixT,
        **options: Unpack[BotOptions],
    ) -> None:
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            bot_id=bot_id,
            **options,
        )

        self._owner_id: str | None = owner_id
        self._get_prefix: PrefixT = prefix
        self._components: dict[str, Component] = {}
        self._base_converter: _BaseConverter = _BaseConverter(self)
        self.__modules: dict[str, types.ModuleType] = {}

    @property
    def bot_id(self) -> str:
        """Property returning the ID of the bot.

        You must ensure you set this via the keyword argument ``bot_id="..."`` in the constructor of this class.

        Returns
        -------
        str
            The ``bot_id`` that was set.
        """
        assert self._bot_id
        return self._bot_id

    @property
    def owner_id(self) -> str | None:
        """Property returning the ID of the user who owns this bot.

        This can be set via the keyword argument ``owner_id="..."`` in the constructor of this class.

        Returns
        -------
        str | None
            The owner ID that has been set. ``None`` if this has not been set.
        """
        return self._owner_id

    async def close(self, **options: Any) -> None:
        for module in tuple(self.__modules):
            try:
                await self.unload_module(module)
            except Exception as e:
                logger.debug('Failed to unload module "%s" gracefully during close: %s.', module, e)

        for component in tuple(self._components):
            try:
                await self.remove_component(component)
            except Exception as e:
                logger.debug('Failed to remove component "%s" gracefully during close: %s.', component, e)

        await super().close(**options)

    def _cleanup_component(self, component: Component, /) -> None:
        for command in component.__all_commands__.values():
            self.remove_command(command.name)

        for listeners in component.__all_listeners__.values():
            for listener in listeners:
                self.remove_listener(listener)

    async def _add_component(self, component: Component, /) -> None:
        for command in component.__all_commands__.values():
            command._injected = component

            if isinstance(command, Group):
                for sub in command.walk_commands():
                    sub._injected = component

            self.add_command(command)

        for name, listeners in component.__all_listeners__.items():
            for listener in listeners:
                self.add_listener(listener, event=name)

        await component.component_load()

    async def add_component(self, component: Component, /) -> None:
        """|coro|

        Method to add a :class:`.commands.Component` to the bot.

        All :class:`~.commands.Command` and :meth:`~.commands.Component.listener`'s in the component will be loaded alongside
        the component.

        If this method fails, including if :meth:`~.commands.Component.component_load` fails, everything will be rolled back
        and cleaned up and a :exc:`.commands.ComponentLoadError` will be raised from the original exception.

        Parameters
        ----------
        component: :class:`~.commands.Component`
            The component to add to the bot.

        Raises
        ------
        ComponentLoadError
            The component failed to load.
        """
        try:
            await self._add_component(component)
        except Exception as e:
            self._cleanup_component(component)
            raise ComponentLoadError from e

        self._components[component.__component_name__] = component

    async def remove_component(self, name: str, /) -> Component | None:
        """|coro|

        Method to remove a :class:`.commands.Component` from the bot.

        All :class:`~.commands.Command` and :meth:`~.commands.Component.listener`'s in the component will be unloaded
        alongside the component.

        If this method fails when :meth:`~.commands.Component.component_teardown` fails, the component will still be unloaded
        completely from the bot, with the exception being logged.

        Parameters
        ----------
        name: str
            The name of the component to unload.

        Returns
        -------
        Component | None
            The component that was removed. ``None`` if the component was not found.
        """
        component: Component | None = self._components.pop(name, None)
        if not component:
            return component

        self._cleanup_component(component)

        try:
            await component.component_teardown()
        except Exception as e:
            msg = f"Ignoring exception in {component.__class__.__qualname__}.component_teardown: {e}\n"
            logger.error(msg, exc_info=e)

        return component

    def get_component(self, name: str, /) -> Component | None:
        """
        Retrieve a Component from the bots loaded Component.
        This will return `None` if the Component was not found.

        Parameters
        ----------
        name: str
            The name of the Component.

        Returns
        -------
        Component | None
        """
        return self._components.get(name)

    def get_context(
        self,
        payload: ChatMessage | ChannelPointsRedemptionAdd | ChannelPointsRedemptionUpdate,
        *,
        cls: Any = None,
    ) -> Any:
        """Method to create and retrieve a :class:`~.commands.Context` object from either a
        :class:`twitchio.ChatMessage`, :class:`twitchio.ChannelPointsRedemptionAdd` or :class:`twitchio.ChannelPointsRedemptionUpdate`.

        This method can be overriden to accept a custom class which inherits from :class:`~.commands.Context` for use within
        the command invocation process, instead of the default.

        Parameters
        ----------
        payload: :class:`~twitchio.ChatMessage` | :class:`twitchio.ChannelPointsRedemptionAdd` | :class:`twitchio.ChannelPointsRedemptionUpdate`
            The message or redemption to construct a :class:`~.commands.Context` or derivative from.
        cls: Any
            A custom class to use as the context. This should inherit from :class:`~.commands.Context`.

        Examples
        --------

        .. code:: python3

            class CustomContext(commands.Context):
                ...

            class Bot(commands.Bot):

                def get_context(self, payload, *, cls: Any = None) -> CustomContext:
                    cls = cls or CustomContext
                    return cls(payload, bot=self)

            @commands.command()
            async def test(ctx: CustomContext) -> None:
                ...
        """
        cls = cls or Context
        return cls(payload, bot=self)

    async def _process_commands(
        self, payload: ChatMessage | ChannelPointsRedemptionAdd | ChannelPointsRedemptionUpdate
    ) -> None:
        ctx: Context = self.get_context(payload)
        await self.invoke(ctx)

    async def process_commands(
        self, payload: ChatMessage | ChannelPointsRedemptionAdd | ChannelPointsRedemptionUpdate
    ) -> None:
        await self._process_commands(payload)

    async def invoke(self, ctx: Context) -> None:
        try:
            await ctx.invoke()
        except CommandError as e:
            payload = CommandErrorPayload(context=ctx, exception=e)
            self.dispatch("command_error", payload=payload)

    async def event_custom_redemption_add(self, payload: ChannelPointsRedemptionAdd) -> None:
        await self.process_commands(payload)

    async def event_custom_redemption_update(self, payload: ChannelPointsRedemptionUpdate) -> None:
        await self.process_commands(payload)

    async def event_message(self, payload: ChatMessage) -> None:
        if payload.chatter.id == self.bot_id:
            return

        if payload.source_broadcaster is not None:
            return

        await self.process_commands(payload)

    async def event_command_error(self, payload: CommandErrorPayload) -> None:
        """An event called when an error occurs during command invocation.

        By default this event logs the exception raised.

        You can override this method, however you should take care to log unhandled exceptions.

        Parameters
        ----------
        payload: :class:`.commands.CommandErrorPayload`
            The payload associated with this event.
        """
        command: Command[Any, ...] | None = payload.context.command
        if command and command.has_error and payload.context.error_dispatched:
            return

        msg = f'Ignoring exception in command "{payload.context.command}":\n'
        logger.error(msg, exc_info=payload.exception)

    async def before_invoke(self, ctx: Context) -> None:
        """A pre invoke hook for all commands that have been added to the bot.

        Commands from :class:`~.commands.Component`'s are included, however if you wish to control them separately,
        see: :meth:`~.commands.Component.component_before_invoke`.

        The pre-invoke hook will be called directly before a valid command is scheduled to run. If this coroutine errors,
        a :exc:`~.commands.CommandHookError` will be raised from the original error.

        Useful for setting up any state like database connections or http clients for command invocation.

        The order of calls with the pre-invoke hooks is:

        - :meth:`.commands.Bot.before_invoke`

        - :meth:`.commands.Component.component_before_invoke`

        - Any ``before_invoke`` hook added specifically to the :class:`~.commands.Command`.


        .. note::

            This hook only runs after successfully parsing arguments and passing all guards associated with the
            command, component (if applicable) and bot.

        Parameters
        ----------
        ctx: :class:`.commands.Context`
            The context associated with command invocation, before being passed to the command.
        """

    async def after_invoke(self, ctx: Context) -> None:
        """A post invoke hook for all commands that have been added to the bot.

        Commands from :class:`~.commands.Component`'s are included, however if you wish to control them separately,
        see: :meth:`~.commands.Component.component_after_invoke`.

        The post-invoke hook will be called after a valid command has completed invocation. If this coroutine errors,
        a :exc:`~.commands.CommandHookError` will be raised from the original error.

        Useful for cleaning up any state like database connections or http clients.

        The order of calls with the post-invoke hooks is:

        - :meth:`.commands.Bot.after_invoke`

        - :meth:`.commands.Component.component_after_invoke`

        - Any ``after_invoke`` hook added specifically to the :class:`~.commands.Command`.


        .. note::

            This hook is always called even when the :class:`~.commands.Command` fails to invoke but similar to
            :meth:`.before_invoke` only if parsing arguments and guards are successfully completed.

        Parameters
        ----------
        ctx: :class:`.commands.Context`
            The context associated with command invocation, after being passed through the command.
        """

    async def global_guard(self, ctx: Context, /) -> bool:
        """|coro|

        A global guard applied to all commmands added to the bot.

        This coroutine function should take in one parameter :class:`~.commands.Context` the context surrounding
        command invocation, and return a bool indicating whether a command should be allowed to run.

        If this function returns ``False``, the chatter will not be able to invoke the command and an error will be
        raised. If this function returns ``True`` the chatter will be able to invoke the command,
        assuming all the other guards also pass their predicate checks.

        See: :func:`~.commands.guard` for more information on guards, what they do and how to use them.

        .. note::

            This is the first guard to run, and is applied to every command.

        .. important::

            Unlike command specific guards or :meth:`.commands.Component.guard`, this function must
            be always be a coroutine.


        This coroutine is intended to be overriden when needed and by default always returns ``True``.

        Parameters
        ----------
        ctx: commands.Context
            The context associated with command invocation.

        Raises
        ------
        GuardFailure
            The guard predicate returned ``False`` and prevented the chatter from using the command.
        """
        return True

    async def subscribe_webhook(
        self,
        payload: SubscriptionPayload,
        *,
        callback_url: str | None = None,
        eventsub_secret: str | None = None,
    ) -> SubscriptionResponse | None:
        return await super().subscribe_webhook(
            payload=payload,
            callback_url=callback_url,
            eventsub_secret=eventsub_secret,
        )

    async def subscribe_websocket(
        self,
        payload: SubscriptionPayload,
        *,
        as_bot: bool | None = None,
        token_for: str | PartialUser | None = None,
        socket_id: str | None = None,
    ) -> SubscriptionResponse | None:
        defaults = payload.default_auth
        if as_bot is None:
            as_bot = defaults.get("as_bot", True)
        return await super().subscribe_websocket(payload=payload, as_bot=as_bot, token_for=token_for, socket_id=socket_id)

    def _get_module_name(self, name: str, package: str | None) -> str:
        try:
            return importlib.util.resolve_name(name, package)
        except ImportError as e:
            raise ModuleNotFoundError(f'The module "{name}" was not found.') from e

    async def _remove_module_remnants(self, name: str) -> None:
        for component_name, component in self._components.copy().items():
            if component.__module__ == name or component.__module__.startswith(f"{name}."):
                await self.remove_component(component_name)

    async def _module_finalizers(self, name: str, module: types.ModuleType) -> None:
        try:
            func = getattr(module, "teardown")
        except AttributeError:
            pass
        else:
            try:
                await func(self)
            except Exception:
                pass
        finally:
            self.__modules.pop(name, None)
            sys.modules.pop(name, None)

            name = module.__name__
            for m in list(sys.modules.keys()):
                if _is_submodule(name, m):
                    del sys.modules[m]

    async def load_module(self, name: str, *, package: str | None = None) -> None:
        """|coro|

        Loads a module.

        A module is a python module that contains commands, cogs, or listeners.

        A module must have a global coroutine, ``setup`` defined as the entry point on what to do when the module is loaded.
        The coroutine takes a single argument, the ``bot``.

        .. versionchanged:: 3.0
            This method is now a :term:`coroutine`.

        Parameters
        ----------
        name: str
            The module to load. It must be dot separated like regular Python imports accessing a sub-module.
            e.g. ``foo.bar`` if you want to import ``foo/bar.py``.
        package: str | None
            The package name to resolve relative imports with.
            This is required when loading an extension using a relative path.
            e.g. ``.foo.bar``. Defaults to ``None``.

        Raises
        ------
        ModuleAlreadyLoadedError
            The module is already loaded.
        ModuleNotFoundError
            The module could not be imported.
            Also raised if module could not be resolved using the `package` parameter.
        ModuleLoadFailure
            There was an error loading the module.
        NoEntryPointError
            The module does not have a setup coroutine.
        TypeError
            The module's setup function is not a coroutine.
        """

        name = self._get_module_name(name, package)

        if name in self.__modules:
            raise ModuleAlreadyLoadedError(f"The module {name} has already been loaded.")

        spec = importlib.util.find_spec(name)
        if spec is None:
            raise ModuleNotFoundError(name)

        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module

        try:
            spec.loader.exec_module(module)  # type: ignore
        except Exception as e:
            del sys.modules[name]
            raise ModuleLoadFailure(name, e) from e

        try:
            entry = getattr(module, "setup")
        except AttributeError as exc:
            del sys.modules[name]
            raise NoEntryPointError(f'The module "{module}" has no setup coroutine.') from exc

        if not asyncio.iscoroutinefunction(entry):
            del sys.modules[name]
            raise TypeError(f'The module "{module}"\'s setup function is not a coroutine.')

        try:
            await entry(self)
        except Exception as e:
            del sys.modules[name]
            await self._remove_module_remnants(module.__name__)
            raise ModuleLoadFailure(name, e) from e

        self.__modules[name] = module

    async def unload_module(self, name: str, *, package: str | None = None) -> None:
        """|coro|

        Unloads a module.

        When the module is unloaded, all commands, listeners and components are removed from the bot, and the module is un-imported.

        You can add an optional global coroutine of ``teardown`` to the module to do miscellaneous clean-up if necessary.
        This also takes a single paramter of the ``bot``, similar to ``setup``.

        .. versionchanged:: 3.0
            This method is now a :term:`coroutine`.

        Parameters
        ----------
        name: str
            The module to unload. It must be dot separated like regular Python imports accessing a sub-module.
            e.g. ``foo.bar`` if you want to import ``foo/bar.py``.
        package: str | None
            The package name to resolve relative imports with.
            This is required when unloading an extension using a relative path.
            e.g. ``.foo.bar``. Defaults to ``None``.

        Raises
        ------
        ModuleNotLoaded
            The module was not loaded.
        """

        name = self._get_module_name(name, package)
        module = self.__modules.get(name)

        if module is None:
            raise ModuleNotLoadedError(name)

        await self._remove_module_remnants(module.__name__)
        await self._module_finalizers(name, module)

    async def reload_module(self, name: str, *, package: str | None = None) -> None:
        """|coro|

        Atomically reloads a module.

        This attempts to unload and then load the module again, in an atomic way.
        If an operation fails mid reload then the bot will revert back to the prior working state.

        .. versionchanged:: 3.0
            This method is now a :term:`coroutine`.

        Parameters
        ----------
        name: str
            The module to unload. It must be dot separated like regular Python imports accessing a sub-module.
            e.g. ``foo.bar`` if you want to import ``foo/bar.py``.
        package: str | None
            The package name to resolve relative imports with.
            This is required when unloading an extension using a relative path.
            e.g. ``.foo.bar``. Defaults to ``None``.

        Raises
        ------
        ModuleNotLoaded
            The module was not loaded.
        ModuleNotFoundError
            The module could not be imported.
            Also raised if module could not be resolved using the `package` parameter.
        ModuleLoadFailure
            There was an error loading the module.
        NoEntryPointError
            The module does not have a setup coroutine.
        TypeError
            The module's setup function is not a coroutine.
        """

        name = self._get_module_name(name, package)
        module = self.__modules.get(name)

        if module is None:
            raise ModuleNotLoadedError(name)

        modules = {name: module for name, module in sys.modules.items() if _is_submodule(module.__name__, name)}

        try:
            await self._remove_module_remnants(module.__name__)
            await self._module_finalizers(name, module)
            await self.load_module(name)
        except Exception as e:
            await module.setup(self)
            self.__modules[name] = module
            sys.modules.update(modules)
            raise e

    @property
    def modules(self) -> Mapping[str, types.ModuleType]:
        """Mapping[:class:`str`, :class:`py:types.ModuleType`]: A read-only mapping of extension name to extension."""
        return types.MappingProxyType(self.__modules)



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, NamedTuple

from twitchio.assets import Asset
from twitchio.user import PartialUser
from twitchio.utils import Colour, parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPAsyncIterator, HTTPClient
    from twitchio.types_.responses import (
        CustomRewardRedemptionResponseData,
        CustomRewardsResponseData,
        CustomRewardsResponseImage,
    )


__all__ = ("CustomReward", "CustomRewardRedemption", "RewardCooldown", "RewardLimitSettings")


class RewardCooldown(NamedTuple):
    """NamedTuple that represents a custom reward's cooldown settings.

    Attributes
    -----------
    enabled: bool
        Whether a coooldown between redemptions is enabled or not. Default is False.
    seconds: int
        The cooldown period in seconds. This only applies if ``is_enabled`` is True.
        Min value is 1; however, the minimum value is 60 for it to be shown in the Twitch UX.
    """

    enabled: bool
    seconds: int


class RewardLimitSettings(NamedTuple):
    """NamedTuple that represents a custom reward's stream limit settings.

    Attributes
    -----------
    enabled: bool
        Whether the stream setting is enabled or not.
    value: int
        The max number of redemptions allowed. Minimum value is 1.
    """

    enabled: bool
    value: int


class CustomReward:
    """Represents a custom reward from a broadcaster's channel.

    Attributes
    -----------
    broadcaster: PartialUser
        The broadcaster that owns the CustomReward.
    id: str
        The ID that uniquely identifies this custom reward.
    title: str
        The title of the reward.
    prompt: str
        The prompt shown to the viewer when they redeem the reward if user input is required.
    cost: int
        The cost of the reward in Channel Points.
    default_image: dict[str, str]
        A dictionary of default images for the reward. The keys are as follows: url_1x, url_2x and url_4x.
    colour: Colour
        The background colour to use for the reward. There is an alias named ``color``.
    enabled: bool
        A Boolean value that determines whether the reward is enabled. Is True if enabled; otherwise, False. Disabled rewards aren't shown to the user.
    input_required: bool
        A Boolean value that determines whether the user must enter information when redeeming the reward. Is True if the reward requires user input.
    paused: bool
        A Boolean value that determines whether the reward is currently paused. Is True if the reward is paused. Viewers can't redeem paused rewards.
    in_stock: bool
        A Boolean value that determines whether the reward is currently in stock. Is True if the reward is in stock. Viewers can't redeem out of stock rewards.
    image: dict[str, str] | None
        A dictionary of custom images for the reward. This will return None if the broadcaster did not upload any images.
        The keys, if available, are as follows: url_1x, url_2x and url_4x.
    skip_queue: bool
        A Boolean value that determines whether redemptions should be set to FULFILLED status immediately when a reward is redeemed. If False, status is set to UNFULFILLED and follows the normal request queue process. The default is False.
    current_stream_redeems: int | None
        The number of redemptions redeemed during the current live stream. The number counts against the ``max_per_stream.max_value`` limit. This is None if the broadcaster's stream isn't live or ``max_per_stream_setting`` isn't enabled.
    cooldown_until: datetime.datetime | None
        The datetime of when the cooldown period expires. Is null if the reward isn't in a cooldown state
    cooldown: RewardCooldown
        The cooldown settings of a reward. This represents whether reward has a cooldown enabled and the cooldown period in seconds.
    max_per_stream: RewardLimitSettings
        The settings of a reward over a live stream. This represents whether a reward has a max number of redemptions per stream and if the setting is enabled or not.
    max_per_user_stream: RewardLimitSettings
        The settings of a reward over a live stream per user. This represents whether a reward has a max number of redemptions per user per stream and if the setting is enabled or not.
    """

    __slots__ = (
        "_http",
        "_image",
        "broadcaster",
        "colour",
        "cooldown",
        "cooldown_until",
        "cost",
        "current_stream_redeems",
        "default_image",
        "enabled",
        "id",
        "in_stock",
        "input_required",
        "max_per_stream",
        "max_per_user_stream",
        "paused",
        "prompt",
        "skip_queue",
        "title",
    )

    def __init__(self, data: CustomRewardsResponseData, *, http: HTTPClient) -> None:
        self._http: HTTPClient = http
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=self._http
        )
        self._image: CustomRewardsResponseImage | None = data.get("image")
        self.id: str = data["id"]
        self.title: str = data["title"]
        self.prompt: str = data["prompt"]
        self.cost: int = int(data["cost"])
        self.default_image: dict[str, str] = {k: str(v) for k, v in data["default_image"].items()}
        self.colour: Colour = Colour.from_hex(data["background_color"])
        self.enabled: bool = data["is_enabled"]
        self.input_required: bool = data["is_user_input_required"]
        self.paused: bool = data["is_paused"]
        self.in_stock: bool = data["is_in_stock"]
        self.skip_queue: bool = data["should_redemptions_skip_request_queue"]
        self.current_stream_redeems: int | None = (
            int(data["redemptions_redeemed_current_stream"])
            if data["redemptions_redeemed_current_stream"] is not None
            else None
        )
        self.cooldown_until: datetime.datetime | None = (
            parse_timestamp(data["cooldown_expires_at"]) if data["cooldown_expires_at"] else None
        )
        self.cooldown: RewardCooldown = RewardCooldown(
            bool(data["global_cooldown_setting"]["is_enabled"]),
            int(data["global_cooldown_setting"]["global_cooldown_seconds"]),
        )
        self.max_per_stream: RewardLimitSettings = RewardLimitSettings(
            bool(data["max_per_stream_setting"]["is_enabled"]), int(data["max_per_stream_setting"]["max_per_stream"])
        )
        self.max_per_user_stream: RewardLimitSettings = RewardLimitSettings(
            bool(data["max_per_user_per_stream_setting"]["is_enabled"]),
            int(data["max_per_user_per_stream_setting"]["max_per_user_per_stream"]),
        )

    def __repr__(self) -> str:
        return f"<CustomReward id={self.id} title={self.title} cost={self.cost}>"

    def __str__(self) -> str:
        return self.title

    @property
    def color(self) -> Colour | None:
        return self.colour

    @property
    def image(self) -> dict[str, str] | None:
        if self._image is not None:
            return {k: str(v) for k, v in self._image.items()}
        else:
            return None

    def get_image(self, size: Literal["1x", "2x", "4x"] = "2x", use_default: bool = False) -> Asset:
        """Get an image Asset for the reward at a specified size.

        Falls back to default images if no custom images have been uploaded.

        Parameters
        ----------
        size: str
            The size key of the image. Options are "1x", "2x", "4x". Defaults to "2x".
        use_default: bool
            Use default images instead of user uploaded images.

        Returns
        -------
        Asset
            The Asset for the image. Falls back to default images if no custom images have been uploaded.
        """
        if use_default or self.image is None or f"url_{size}" not in self.image:
            url = self.default_image[f"url_{size}"]
        else:
            url = self.image[f"url_{size}"]

        return Asset(url, http=self._http)

    async def delete(self) -> None:
        """|coro|

        Delete the custom reward.

        The app / client ID used to create the reward is the only app that may delete it.
        If the reward's redemption status is UNFULFILLED at the time the reward is deleted, its redemption status is marked as FULFILLED.

        .. note::
            Requires a user access token that includes the ``channel:manage:redemptions`` scope.
        """
        await self._http.delete_custom_reward(
            broadcaster_id=self.broadcaster.id, reward_id=self.id, token_for=self.broadcaster.id
        )

    async def update(
        self,
        *,
        title: str | None = None,
        cost: int | None = None,
        prompt: str | None = None,
        enabled: bool | None = None,
        colour: str | Colour | None = None,
        input_required: bool | None = None,
        max_per_stream: int | None = None,
        max_per_user: int | None = None,
        global_cooldown: int | None = None,
        paused: bool | None = None,
        skip_queue: bool | None = None,
    ) -> CustomReward:
        """|coro|

        Update the custom reward.

        .. important::
            The app / client ID used to create the reward is the only app that may update the reward.

        .. note::
            Requires a user access token that includes the ``channel:manage:redemptions`` scope.

        Parameters
        -----------
        title: str | None
            The reward's title.
            The title may contain a maximum of 45 characters and it must be unique amongst all of the broadcaster's custom rewards.
        cost: int | None
            The cost of the reward, in channel points. The minimum is 1 point.
        prompt: str | None
            The prompt shown to the viewer when they redeem the reward.
            ``input_required`` needs to be set to ``True`` for this to work,
        enabled: bool | None
             Boolean value that indicates whether the reward is enabled. Set to ``True`` to enable the reward. Viewers see only enabled rewards.
        colour: str | Colour | None
            The background colour to use for the reward. Specify the color using Hex format (for example, #00E5CB).
            You can also pass a twitchio.Colour object.
        input_required: bool | None
            A Boolean value that determines whether users must enter information to redeem the reward.
        max_per_stream: int | None
            The maximum number of redemptions allowed per live stream.
            Setting this to 0 disables the maximum number of redemptions per stream.
        max_per_user: int | None
            The maximum number of redemptions allowed per user per live stream.
            Setting this to 0 disables the maximum number of redemptions per user per stream.
        global_cooldown: int | None
            The cooldown period, in seconds. The minimum value is 1; however, for it to be shown in the Twitch UX, the minimum value is 60.
            Setting this to 0 disables the global cooldown period.
        paused: bool | None
            A Boolean value that determines whether to pause the reward. Set to ``True`` to pause the reward. Viewers can't redeem paused rewards.
        skip_queue: bool | None
            A Boolean value that determines whether redemptions should be set to FULFILLED status immediately when a reward is redeemed.
            If False, status is set to UNFULFILLED and follows the normal request queue process.

        Returns
        --------
        CustomReward

        Raises
        ------
        ValueError
            title must be a maximum of 45 characters.
        ValueError
            prompt must be a maximum of 200 characters.
        ValueError
            Minimum value must be at least 1.
        """

        if title is not None and len(title) > 45:
            raise ValueError("title must be a maximum of 45 characters.")
        if cost is not None and cost < 1:
            raise ValueError("cost must be at least 1.")
        if prompt is not None and len(prompt) > 200:
            raise ValueError("prompt must be a maximum of 200 characters.")

        data = await self._http.patch_custom_reward(
            broadcaster_id=self.broadcaster.id,
            token_for=self.broadcaster.id,
            reward_id=self.id,
            title=title,
            cost=cost,
            prompt=prompt,
            enabled=enabled,
            background_color=colour,
            user_input_required=input_required,
            max_per_stream=max_per_stream,
            max_per_user=max_per_user,
            global_cooldown=global_cooldown,
            paused=paused,
            skip_queue=skip_queue,
        )

        return CustomReward(data=data["data"][0], http=self._http)

    def fetch_redemptions(
        self,
        *,
        status: Literal["CANCELED", "FULFILLED", "UNFULFILLED"],
        ids: list[str] | None = None,
        sort: Literal["OLDEST", "NEWEST"] = "OLDEST",
        first: int = 20,
    ) -> HTTPAsyncIterator[CustomRewardRedemption]:
        """|aiter|

        Fetch redemptions from the CustomReward.

        Canceled and fulfilled redemptions are returned for only a few days after they're canceled or fulfilled.

        .. note::
            Requires a user access token that includes the ``channel:read:redemptions`` or ``channel:manage:redemptions`` scope.

        Parameters
        -----------
        status: typing.Literal["CANCELED", "FULFILLED", "UNFULFILLED"]
            The state of the redemption. This can be one of the following: "CANCELED", "FULFILLED", "UNFULFILLED"
        ids: list[str] | None
            A list of IDs to filter the redemptions by. You may specify up to 50.
        sort: typing.Literal["OLDEST", "NEWEST"]
            The order to sort the redemptions by. The default is OLDEST.
        first: int
            Maximum number of items to return per page. Default is 20.
            Min is 1 and Max is 50.

        Returns
        --------
        HTTPAsyncIterator[CustomRewardRedemption]

        Raises
        ------
        ValueError
            You may only specify up to 50 redemption ids.
        """

        first = max(1, min(50, first))

        if ids is not None and len(ids) > 50:
            raise ValueError("You may only specify up to 50 redemption ids.")

        return self._http.get_custom_reward_redemptions(
            broadcaster_id=self.broadcaster.id,
            token_for=self.broadcaster.id,
            reward_id=self.id,
            ids=ids,
            status=status,
            sort=sort,
            first=first,
            parent_reward=self,
        )


class CustomRewardRedemption:
    """
    Represents a custom reward redemption.

    Attributes
    -----------
    id: str
        The ID that uniquely identifies this redemption.
    status: typing.Literal["CANCELED", "FULFILLED", "UNFULFILLED"]
        The state of the redemption. This can be one of the following: "CANCELED", "FULFILLED", "UNFULFILLED"
    redeemed_at: datetime.datetime
        The prompt shown to the viewer when they redeem the reward if user input is required.
    reward: CustomReward
        This is the reward that the redemption is from.
    user: PartialUser
        The user that made the redemption.
    """

    __slots__ = ("_http", "id", "redeemed_at", "reward", "status", "user")

    def __init__(
        self,
        data: CustomRewardRedemptionResponseData,
        parent_reward: CustomReward,
        http: HTTPClient,
    ) -> None:
        self.id = data["id"]
        self.status: Literal["CANCELED", "FULFILLED", "UNFULFILLED"] = data["status"]
        self.redeemed_at: datetime.datetime = parse_timestamp(data["redeemed_at"])
        self.reward: CustomReward = parent_reward
        self._http: HTTPClient = http
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], data["broadcaster_name"], http=self._http)

    def __repr__(self) -> str:
        return f"<CustomRewardRedemption id={self.id} status={self.status} redeemed_at={self.redeemed_at}>"

    async def fulfill(self) -> CustomRewardRedemption:
        """|coro|

        Updates a redemption's status to FULFILLED.

        .. note::
            Requires a user access token that includes the ``channel:manage:redemptions`` scope.


        Returns
        --------
        CustomRewardRedemption
        """
        data = await self._http.patch_custom_reward_redemption(
            broadcaster_id=self.reward.broadcaster.id,
            id=self.reward.broadcaster.id,
            token_for=self,
            reward_id=self.reward.id,
            status="FULFILLED",
        )
        return CustomRewardRedemption(data["data"][0], parent_reward=self.reward, http=self._http)

    async def refund(self) -> CustomRewardRedemption:
        """|coro|

        Updates a redemption's status to CANCELED.

        .. note::
            Requires a user access token that includes the ``channel:manage:redemptions`` scope.

        Returns
        --------
        CustomRewardRedemption
        """
        data = await self._http.patch_custom_reward_redemption(
            broadcaster_id=self.reward.broadcaster.id,
            id=self.id,
            token_for=self.reward.broadcaster.id,
            reward_id=self.reward.id,
            status="CANCELED",
        )
        return CustomRewardRedemption(data["data"][0], parent_reward=self.reward, http=self._http)



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from twitchio.assets import Asset
from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp

from .games import Game


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import (
        CreateStreamMarkerResponseData,
        GamesResponse,
        StreamMarkersResponseData,
        StreamMarkersResponseMarkers,
        StreamsResponseData,
    )


__all__ = ("Stream", "StreamMarker", "VideoMarkers")


class Stream:
    """Represents a Stream

    Attributes
    -----------
    id: str
        The current stream ID.
    user: twitchio.PartialUser
        The user who is streaming.
    game_id: str | None
        Current game ID being played on the channel. Could be `None` if no category / game has been set.
    game_name: str | None
        Name of the game being played on the channel. Could be `None` if no category / game has been set.
    type: str
        Whether the stream is "live" or not.
    title: str
        Title of the stream.
    viewer_count: int
        Current viewer count of the stream
    started_at: datetime.datetime
        UTC timestamp of when the stream started.
    language: str
        Language of the channel.
    thumbnail: Asset
        The :class:`~twitchio.Asset` which can be used to read or save the thumbnail associated with this stream.
    is_mature: bool
        Indicates whether the stream is intended for mature audience.
    tags: list[str]
        The tags applied to the channel.
    """

    __slots__ = (
        "_http",
        "game_id",
        "game_name",
        "id",
        "is_mature",
        "language",
        "started_at",
        "tags",
        "thumbnail",
        "title",
        "type",
        "user",
        "viewer_count",
    )

    def __init__(self, data: StreamsResponseData, *, http: HTTPClient) -> None:
        self._http: HTTPClient = http

        self.id: str = data["id"]
        self.user = PartialUser(data["user_id"], data["user_login"], data["user_name"], http=http)
        self.game_id: str | None = data["game_id"] or None
        self.game_name: str | None = data["game_name"] or None
        self.type: str = data["type"]
        self.title: str = data["title"]
        self.viewer_count: int = data["viewer_count"]
        self.started_at = parse_timestamp(data["started_at"])
        self.language: str = data["language"]
        self.thumbnail: Asset = Asset(data["thumbnail_url"], dimensions=(640, 360), http=http)
        self.is_mature: bool = data["is_mature"]
        self.tags: list[str] = data["tags"]

    def __repr__(self) -> str:
        return f"<Stream id={self.id} user={self.user} title={self.title} started_at={self.started_at}>"

    async def fetch_game(self) -> Game | None:
        """|coro|

        Fetches the :class:`~twitchio.Game` associated with this stream.

        The :class:`~twitchio.Game` returned is current from the time the :class:`~twitchio.Stream`
        instance was created.

        Could be `None` if no category / game was set at the time the :class:`~twitchio.Stream`
        instance was created.

        Returns
        -------
        twitchio.Game | None
            The game associated with this :class:`~twitchio.Stream` instance, or `None`.
        """
        if self.game_id is None:
            return None

        payload: GamesResponse = await self._http.get_games(ids=[self.game_id])
        return Game(payload["data"][0], http=self._http) if payload["data"] else None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stream):
            return NotImplemented

        return self.id == other.id and self.started_at == other.started_at


class StreamMarker:
    """Represents a stream marker.

    Attributes
    ----------
    id: str
        An ID that identifies the single market you added.
    created_at: datetime.datetime
        Datetime of when the user created the stream marker.
    position: int
        The relative offset (in seconds) of the marker from the beginning of the stream.
    url: str | None
        A URL that opens the video in Twitch Highlighter. This is None on creation but populated when fetched.
    """

    __slots__ = ("created_at", "description", "id", "position", "url")

    def __init__(self, data: CreateStreamMarkerResponseData | StreamMarkersResponseMarkers) -> None:
        self.id: str = data["id"]
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.description: str = data["description"]
        self.position: int = int(data["position_seconds"])
        self.url: str | None = data.get("url")

    def __repr__(self) -> str:
        return f"<StreamMarker id={self.id} created_at={self.created_at} position={self.position}>"


class VideoMarkers:
    """Represents stream markers for the latest stream.

    Attributes
    ----------
    user: PartialUser
        The user that created the stream marker.
    video_id: str
        An ID that identifies this video.
    markers: list[StreamMarker]
        The list of markers in this video. The list in ascending order by when the marker was created.
    """

    __slots__ = ("markers", "user", "video_id")

    def __init__(self, data: StreamMarkersResponseData, *, http: HTTPClient) -> None:
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], data["user_name"], http=http)
        self.video_id: str = data["videos"][0]["video_id"]
        self.markers: list[StreamMarker] = [StreamMarker(d) for d in data["videos"][0]["markers"]]

    def __repr__(self) -> str:
        return f"<VideoMarkers user={self.user} video_id={self.video_id}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Any, Generic, Literal, NotRequired, TypeAlias, TypedDict, TypeVar

from .conduits import Condition, ConduitData, ShardData


__all__ = (
    "AdScheduleResponse",
    "AdScheduleResponseData",
    "AuthorizationURLResponse",
    "BitsLeaderboardResponse",
    "ChannelChatBadgesResponseData",
    "ChannelChatBadgesResponseVersions",
    "ChannelEditorsResponse",
    "ChannelEditorsResponse",
    "ChannelEditorsResponseData",
    "ChannelEmotesResponse",
    "ChannelFollowersResponseData",
    "ChannelInformationResponse",
    "ChannelInformationResponseData",
    "ChannelTeamsResponseData",
    "CheermotesResponse",
    "CheermotesResponseData",
    "CheermotesResponseTiers",
    "ClientCredentialsResponse",
    "ClipsResponseData",
    "ConduitPayload",
    "ContentClassificationLabelData",
    "ContentClassificationLabelsResponse",
    "EventsubSubscriptionResponse",
    "EventsubSubscriptionResponseData",
    "ExtensionAnalyticsResponse",
    "ExtensionTransactionsResponse",
    "ExtensionTransactionsResponseCost",
    "ExtensionTransactionsResponseData",
    "ExtensionTransactionsResponseProductData",
    "FollowedChannelsResponseData",
    "GameAnalyticsResponse",
    "GamesResponse",
    "GamesResponseData",
    "GlobalChatBadgesResponse",
    "GlobalChatBadgesResponseData",
    "GlobalChatBadgesResponseVersions",
    "GlobalEmotesResponse",
    "GlobalEmotesResponseData",
    "OAuthResponses",
    "RawResponse",
    "RefreshTokenResponse",
    "SearchChannelsResponseData",
    "ShardPayload",
    "SharedChatSessionResponse",
    "SharedChatSessionResponseData",
    "SnoozeNextAdResponse",
    "StartCommercialResponse",
    "StartCommercialResponseData",
    "StreamsResponseData",
    "TeamsResponseData",
    "TopGamesResponseData",
    "UserChatColorResponse",
    "UserChatColorResponseData",
    "UserTokenResponse",
    "ValidateTokenResponse",
    "VideosResponseData",
    "WarnChatUserResponse",
    "WarnChatUserResponseData",
)

T = TypeVar("T")


class Payload(TypedDict, Generic[T]):
    data: list[T]


class _TokenResponseBase(TypedDict):
    access_token: str
    refresh_token: str
    expires_in: int
    scope: str | list[str]
    token_type: str


RefreshTokenResponse: TypeAlias = _TokenResponseBase
UserTokenResponse: TypeAlias = _TokenResponseBase


class ValidateTokenResponse(TypedDict):
    client_id: str
    login: str
    scopes: list[str]
    user_id: str
    expires_in: int


class ClientCredentialsResponse(TypedDict):
    access_token: str
    expires_in: int
    token_type: str


class AuthorizationURLResponse(TypedDict):
    url: str
    client_id: str
    redirect_uri: str
    response_type: str
    scopes: list[str]
    force_verify: bool
    state: str


OAuthResponses: TypeAlias = (
    RefreshTokenResponse | ValidateTokenResponse | ClientCredentialsResponse | UserTokenResponse | AuthorizationURLResponse
)
RawResponse: TypeAlias = dict[str, Any]


class Pagination(TypedDict):
    cursor: str | None


class DateRange(TypedDict):
    started_at: str
    ended_at: str


class StartCommercialResponseData(TypedDict):
    length: int
    message: str
    retry_after: int


class StartCommercialResponse(TypedDict):
    data: list[StartCommercialResponseData]


class AdScheduleResponseData(TypedDict):
    snooze_count: int
    snooze_refresh_at: str
    next_ad_at: str
    duration: int
    last_ad_at: str
    preroll_free_time: int


class AdScheduleResponse(TypedDict):
    data: list[AdScheduleResponseData]


class SnoozeNextAdResponseData(TypedDict):
    snooze_count: int
    snooze_refresh_at: str
    next_ad_at: str


class SnoozeNextAdResponse(TypedDict):
    data: list[SnoozeNextAdResponseData]


class ExtensionAnalyticsResponseData(TypedDict):
    extension_id: str
    URL: str
    type: str
    date_range: DateRange


class ExtensionAnalyticsResponse(TypedDict):
    data: list[ExtensionAnalyticsResponseData]
    pagination: Pagination


class GameAnalyticsResponseData(TypedDict):
    game_id: str
    URL: str
    type: str
    date_range: DateRange


class GameAnalyticsResponse(TypedDict):
    data: list[GameAnalyticsResponseData]
    pagination: Pagination


class BitsLeaderboardResponseData(TypedDict):
    user_id: str
    user_login: str
    user_name: str
    rank: int
    score: int


class BitsLeaderboardResponse(TypedDict):
    data: list[BitsLeaderboardResponseData]
    date_range: DateRange
    total: int


class CheermotesResponseTiers(TypedDict):
    min_bits: int
    id: str
    color: str
    images: dict[str, dict[str, dict[str, str]]]
    can_cheer: bool
    show_in_bits_card: bool


class CheermotesResponseData(TypedDict):
    prefix: str
    tiers: list[CheermotesResponseTiers]
    type: str
    order: int
    last_updated: str
    is_charitable: bool


class CheermotesResponse(TypedDict):
    data: list[CheermotesResponseData]


class ExtensionTransactionsResponseCost(TypedDict):
    amount: int
    type: str


class ExtensionTransactionsResponseProductData(TypedDict):
    sku: str
    domain: str
    cost: ExtensionTransactionsResponseCost
    inDevelopment: bool
    displayName: str
    expiration: str
    broadcast: bool


class ExtensionTransactionsResponseData(TypedDict):
    id: str
    timestamp: str
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    user_id: str
    user_login: str
    user_name: str
    product_type: str
    product_data: ExtensionTransactionsResponseProductData


class ExtensionTransactionsResponse(TypedDict):
    data: list[ExtensionTransactionsResponseData]
    pagination: Pagination


class ChannelInformationResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    broadcaster_language: str
    game_name: str
    game_id: str
    title: str
    delay: int
    tags: list[str]
    content_classification_labels: list[str]
    is_branded_content: bool


class ChannelInformationResponse(TypedDict):
    data: list[ChannelInformationResponseData]


class ChannelEditorsResponseData(TypedDict):
    user_id: str
    user_name: str
    created_at: str


class ChannelEditorsResponse(TypedDict):
    data: list[ChannelEditorsResponseData]


class FollowedChannelsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    followed_at: str


class FollowedChannelsResponse(TypedDict):
    data: list[FollowedChannelsResponseData]
    pagination: Pagination
    total: int


class ChannelFollowersResponseData(TypedDict):
    followed_at: str
    user_id: str
    user_login: str
    user_name: str


class ChannelFollowersResponse(TypedDict):
    data: list[ChannelFollowersResponseData]
    pagination: Pagination
    total: int


class CustomRewardsResponseImage(TypedDict):
    url_1x: str
    url_2x: str
    url_4x: str


class CustomRewardsResponseDefaultImage(TypedDict):
    url_1x: str
    url_2x: str
    url_4x: str


class CustomRewardsResponseMaxPerStreamSetting(TypedDict):
    is_enabled: bool
    max_per_stream: int


class CustomRewardsResponseMaxPerUserPerStreamSetting(TypedDict):
    is_enabled: bool
    max_per_user_per_stream: int


class CustomRewardsResponseGlobalCooldownSetting(TypedDict):
    is_enabled: bool
    global_cooldown_seconds: int


class CustomRewardsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    id: str
    title: str
    prompt: str
    cost: int
    image: CustomRewardsResponseImage | None
    default_image: CustomRewardsResponseDefaultImage
    background_color: str
    is_enabled: bool
    is_user_input_required: bool
    max_per_stream_setting: CustomRewardsResponseMaxPerStreamSetting
    max_per_user_per_stream_setting: CustomRewardsResponseMaxPerUserPerStreamSetting
    global_cooldown_setting: CustomRewardsResponseGlobalCooldownSetting
    is_paused: bool
    is_in_stock: bool
    should_redemptions_skip_request_queue: bool
    redemptions_redeemed_current_stream: int | None
    cooldown_expires_at: str | None


class CustomRewardsResponse(TypedDict):
    data: list[CustomRewardsResponseData]


class CustomRewardRedemptionResponseReward(TypedDict):
    id: str
    title: str
    prompt: str
    cost: int


class CustomRewardRedemptionResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    id: str
    user_login: str
    user_id: str
    user_name: str
    user_input: str
    status: Literal["CANCELED", "FULFILLED", "UNFULFILLED"]
    redeemed_at: str
    reward: CustomRewardRedemptionResponseReward


class CustomRewardRedemptionResponse(TypedDict):
    data: list[CustomRewardRedemptionResponseData]


class CharityCampaignResponseCurrentAmount(TypedDict):
    value: int
    decimal_places: int
    currency: str


class CharityCampaignResponseTargetAmount(TypedDict):
    value: int
    decimal_places: int
    currency: str


class CharityCampaignResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    charity_name: str
    charity_description: str
    charity_logo: str
    charity_website: str
    current_amount: CharityCampaignResponseCurrentAmount
    target_amount: CharityCampaignResponseTargetAmount | None


class CharityCampaignResponse(TypedDict):
    data: list[CharityCampaignResponseData]


class CharityCampaignDonationsResponseAmount(TypedDict):
    value: int
    decimal_places: int
    currency: str


class CharityCampaignDonationsResponseData(TypedDict):
    id: str
    campaign_id: str
    user_id: str
    user_login: str
    user_name: str
    amount: CharityCampaignDonationsResponseAmount


class CharityCampaignDonationsResponse(TypedDict):
    data: list[CharityCampaignDonationsResponseData]
    pagination: Pagination


class ChattersResponseData(TypedDict):
    user_id: str
    user_login: str
    user_name: str


class ChattersResponse(TypedDict):
    data: list[ChattersResponseData]
    pagination: Pagination
    total: int


class ChannelEmotesResponseImages(TypedDict):
    url_1x: str
    url_2x: str
    url_4x: str


class ChannelEmotesResponseData(TypedDict):
    id: str
    name: str
    images: ChannelEmotesResponseImages
    tier: Literal["1000", "2000", "3000"]
    emote_type: Literal["bitstier", "follower", "subscriptions"]
    emote_set_id: str
    format: list[str]
    scale: list[str]
    theme_mode: list[str]


class ChannelEmotesResponse(TypedDict):
    data: list[ChannelEmotesResponseData]
    template: str


class GlobalEmotesResponseImages(TypedDict):
    url_1x: str
    url_2x: str
    url_4x: str


class GlobalEmotesResponseData(TypedDict):
    id: str
    name: str
    images: GlobalEmotesResponseImages
    format: list[str]
    scale: list[str]
    theme_mode: list[str]


class GlobalEmotesResponse(TypedDict):
    data: list[GlobalEmotesResponseData]
    template: str


class EmoteSetsResponseImages(TypedDict):
    url_1x: str
    url_2x: str
    url_4x: str


class EmoteSetsResponseData(TypedDict):
    id: str
    name: str
    images: EmoteSetsResponseImages
    emote_type: str
    emote_set_id: str
    owner_id: str
    format: list[str]
    scale: list[str]
    theme_mode: list[str]


class EmoteSetsResponse(TypedDict):
    data: list[EmoteSetsResponseData]
    template: str


class ChannelChatBadgesResponseVersions(TypedDict):
    id: str
    image_url_1x: str
    image_url_2x: str
    image_url_4x: str
    title: str
    description: str
    click_action: str | None
    click_url: str | None


class ChannelChatBadgesResponseData(TypedDict):
    set_id: str
    versions: list[ChannelChatBadgesResponseVersions]


class ChannelChatBadgesResponse(TypedDict):
    data: list[ChannelChatBadgesResponseData]


class GlobalChatBadgesResponseVersions(TypedDict):
    id: str
    image_url_1x: str
    image_url_2x: str
    image_url_4x: str
    title: str
    description: str
    click_action: str | None
    click_url: str | None


class GlobalChatBadgesResponseData(TypedDict):
    set_id: str
    versions: list[GlobalChatBadgesResponseVersions]


class GlobalChatBadgesResponse(TypedDict):
    data: list[GlobalChatBadgesResponseData]


class ChatSettingsResponseData(TypedDict):
    broadcaster_id: str
    emote_mode: bool
    follower_mode: bool
    follower_mode_duration: int | None
    moderator_id: str
    non_moderator_chat_delay: bool
    non_moderator_chat_delay_duration: int | None
    slow_mode: bool
    slow_mode_wait_time: int | None
    subscriber_mode: bool
    unique_chat_mode: bool


class ChatSettingsResponse(TypedDict):
    data: list[ChatSettingsResponseData]


class SharedChatParticipantsData(TypedDict):
    broadcaster_id: str


class SharedChatSessionResponseData(TypedDict):
    session_id: str
    host_broadcaster_id: str
    participants: list[SharedChatParticipantsData]
    created_at: str
    updated_at: str


class SharedChatSessionResponse(TypedDict):
    data: list[SharedChatSessionResponseData]


class UserEmotesResponseData(TypedDict):
    id: str
    name: str
    emote_type: str
    emote_set_id: str
    owner_id: str
    format: list[str]
    scale: list[str]
    theme_mode: list[str]


class UserEmotesResponse(TypedDict):
    data: list[UserEmotesResponseData]
    template: str
    pagination: Pagination


class SendChatMessageResponseDropReason(TypedDict):
    code: str
    message: str


class SendChatMessageResponseData(TypedDict):
    message_id: str
    is_sent: bool
    drop_reason: SendChatMessageResponseDropReason | None


class SendChatMessageResponse(TypedDict):
    data: list[SendChatMessageResponseData]


class UserChatColorResponseData(TypedDict):
    user_id: str
    user_login: str
    user_name: str
    color: str


class UserChatColorResponse(TypedDict):
    data: list[UserChatColorResponseData]


class CreateClipResponseData(TypedDict):
    edit_url: str
    id: str


class CreateClipResponse(TypedDict):
    data: list[CreateClipResponseData]


class ClipsResponseData(TypedDict):
    id: str
    url: str
    embed_url: str
    broadcaster_id: str
    broadcaster_name: str
    creator_id: str
    creator_name: str
    video_id: str
    game_id: str
    language: str
    title: str
    view_count: int
    created_at: str
    thumbnail_url: str
    duration: float
    vod_offset: int | None
    is_featured: bool


class ClipsResponse(TypedDict):
    data: list[ClipsResponseData]
    pagination: Pagination


class ConduitPayload(TypedDict):
    data: list[ConduitData]


class ShardPayload(TypedDict):
    data: list[ShardData]


class ContentClassificationLabelData(TypedDict):
    id: str
    description: str
    name: str


class ContentClassificationLabelsResponse(TypedDict):
    data: list[ContentClassificationLabelData]


class DropsEntitlementsResponseData(TypedDict):
    id: str
    benefit_id: str
    timestamp: str
    user_id: str
    game_id: str
    fulfillment_status: Literal["CLAIMED", "FULFILLED"]
    last_updated: str


class DropsEntitlementsResponse(TypedDict):
    data: list[DropsEntitlementsResponseData]
    pagination: Pagination


class UpdateDropsEntitlementsResponseData(TypedDict):
    status: Literal["INVALID_ID", "NOT_FOUND", "SUCCESS", "UNAUTHORIZED", "UPDATE_FAILED"]
    ids: list[str]


class UpdateDropsEntitlementsResponse(TypedDict):
    data: list[UpdateDropsEntitlementsResponseData]


class ExtensionConfigurationSegmentResponseData(TypedDict):
    segment: str
    broadcaster_id: str
    content: str
    version: str


class ExtensionConfigurationSegmentResponse(TypedDict):
    data: list[ExtensionConfigurationSegmentResponseData]


class ExtensionLiveChannelsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_name: str
    game_name: str
    game_id: str
    title: str


class ExtensionLiveChannelsResponse(TypedDict):
    data: list[ExtensionLiveChannelsResponseData]
    pagination: Pagination


class ExtensionSecretsResponseSecrets(TypedDict):
    content: str
    active_at: str
    expires_at: str


class ExtensionSecretsResponseData(TypedDict):
    format_version: int
    secrets: list[ExtensionSecretsResponseSecrets]


class ExtensionSecretsResponse(TypedDict):
    data: list[ExtensionSecretsResponseData]


class CreateExtensionSecretResponseSecrets(TypedDict):
    content: str
    active_at: str
    expires_at: str


class CreateExtensionSecretResponseData(TypedDict):
    format_version: int
    secrets: list[CreateExtensionSecretResponseSecrets]


class CreateExtensionSecretResponse(TypedDict):
    data: list[CreateExtensionSecretResponseData]


class ExtensionsResponseMobile(TypedDict):
    viewer_url: str


class ExtensionsResponsePanel(TypedDict):
    viewer_url: str
    height: int
    can_link_external_content: bool


class ExtensionsResponseVideoOverlay(TypedDict):
    viewer_url: str
    can_link_external_content: bool


class ExtensionsResponseComponent(TypedDict):
    viewer_url: str
    aspect_ratio_x: int
    aspect_ratio_y: int
    autoscale: bool
    scale_pixels: int
    target_height: int
    can_link_external_content: bool


class ExtensionsResponseConfig(TypedDict):
    viewer_url: str
    can_link_external_content: bool


class ExtensionsResponseViews(TypedDict):
    mobile: ExtensionsResponseMobile
    panel: ExtensionsResponsePanel
    video_overlay: ExtensionsResponseVideoOverlay
    component: ExtensionsResponseComponent
    config: ExtensionsResponseConfig


class ExtensionsResponseData(TypedDict):
    author_name: str
    bits_enabled: bool
    can_install: bool
    configuration_location: str
    description: str
    eula_tos_url: str
    has_chat_support: bool
    icon_url: str
    icon_urls: dict[str, str]
    id: str
    name: str
    privacy_policy_url: str
    request_identity_link: bool
    screenshot_urls: list[str]
    state: str
    subscriptions_support_level: str
    summary: str
    support_email: str
    version: str
    viewer_summary: str
    views: ExtensionsResponseViews
    allowlisted_config_urls: list[str]
    allowlisted_panel_urls: list[str]


class ExtensionsResponse(TypedDict):
    data: list[ExtensionsResponseData]


class ReleasedExtensionsResponseMobile(TypedDict):
    viewer_url: str


class ReleasedExtensionsResponsePanel(TypedDict):
    viewer_url: str
    height: int
    can_link_external_content: bool


class ReleasedExtensionsResponseVideoOverlay(TypedDict):
    viewer_url: str
    can_link_external_content: bool


class ReleasedExtensionsResponseComponent(TypedDict):
    viewer_url: str
    aspect_ratio_x: int
    aspect_ratio_y: int
    autoscale: bool
    scale_pixels: int
    target_height: int
    can_link_external_content: bool


class ReleasedExtensionsResponseConfig(TypedDict):
    viewer_url: str
    can_link_external_content: bool


class ReleasedExtensionsResponseViews(TypedDict):
    mobile: ReleasedExtensionsResponseMobile
    panel: ReleasedExtensionsResponsePanel
    video_overlay: ReleasedExtensionsResponseVideoOverlay
    component: ReleasedExtensionsResponseComponent
    config: ReleasedExtensionsResponseConfig


class ReleasedExtensionsResponseData(TypedDict):
    author_name: str
    bits_enabled: bool
    can_install: bool
    configuration_location: str
    description: str
    eula_tos_url: str
    has_chat_support: bool
    icon_url: str
    icon_urls: dict[str, str]
    id: str
    name: str
    privacy_policy_url: str
    request_identity_link: bool
    screenshot_urls: list[str]
    state: str
    subscriptions_support_level: str
    summary: str
    support_email: str
    version: str
    viewer_summary: str
    views: ReleasedExtensionsResponseViews
    allowlisted_config_urls: list[str]
    allowlisted_panel_urls: list[str]


class ReleasedExtensionsResponse(TypedDict):
    data: list[ReleasedExtensionsResponseData]


class ExtensionBitsProductsResponseCost(TypedDict):
    amount: int
    type: str


class ExtensionBitsProductsResponseData(TypedDict):
    sku: str
    cost: ExtensionBitsProductsResponseCost
    in_development: bool
    display_name: str
    expiration: str
    is_broadcast: bool


class ExtensionBitsProductsResponse(TypedDict):
    data: list[ExtensionBitsProductsResponseData]


class UpdateExtensionBitsProductResponseCost(TypedDict):
    amount: int
    type: str


class UpdateExtensionBitsProductResponseData(TypedDict):
    sku: str
    cost: UpdateExtensionBitsProductResponseCost
    in_development: bool
    display_name: str
    expiration: str
    is_broadcast: bool


class UpdateExtensionBitsProductResponse(TypedDict):
    data: list[UpdateExtensionBitsProductResponseData]


class EventsubTransportData(TypedDict):
    method: Literal["websocket", "webhook"]
    callback: NotRequired[str]
    session_id: NotRequired[str]
    connected_at: NotRequired[str]
    disconnected_at: NotRequired[str]


class EventsubSubscriptionResponseData(TypedDict):
    id: str
    status: Literal[
        "enabled",
        "webhook_callback_verification_pending",
        "webhook_callback_verification_failed",
        "notification_failures_exceeded",
        "authorization_revoked",
        "moderator_removed",
        "user_removed",
        "version_removed",
        "beta_maintenance",
        "websocket_disconnected",
        "websocket_failed_ping_pong",
        "websocket_received_inbound_traffic",
        "websocket_connection_unused",
        "websocket_internal_error",
        "websocket_network_timeout",
        "websocket_network_error",
    ]
    type: str
    version: str
    condition: Condition
    created_at: str
    cost: int
    transport: EventsubTransportData


class EventsubSubscriptionResponse(TypedDict):
    data: list[EventsubSubscriptionResponseData]
    total: int
    total_cost: int
    max_total_cost: int
    pagination: Pagination


class TopGamesResponseData(TypedDict):
    id: str
    name: str
    box_art_url: str
    igdb_id: str


class TopGamesResponse(TypedDict):
    data: list[TopGamesResponseData]
    pagination: Pagination


class GamesResponseData(TypedDict):
    id: str
    name: str
    box_art_url: str
    igdb_id: str


class GamesResponse(TypedDict):
    data: list[GamesResponseData]


class CreatorGoalsResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str
    type: Literal["follower", "subscription", "subscription_count", "new_subscription", "new_subscription_count"]
    description: str
    current_amount: int
    target_amount: int
    created_at: str


class CreatorGoalsResponse(TypedDict):
    data: list[CreatorGoalsResponseData]


class ChannelGuestStarSettingsResponse(TypedDict):
    is_moderator_send_live_enabled: bool
    slot_count: int
    is_browser_source_audio_enabled: bool
    group_layout: str
    browser_source_token: str


# TODO GUESTS
class GuestStarSessionResponse(TypedDict):
    id: str


class CreateGuestStarSessionResponse(TypedDict): ...


class EndGuestStarSessionResponse(TypedDict): ...


class GuestStarInvitesResponse(TypedDict): ...


class HypeTrainEventsResponseContributions(TypedDict):
    total: int
    type: Literal["BITS", "SUBS", "OTHER"]
    user: str


class HypeTrainEventsResponseEventData(TypedDict):
    broadcaster_id: str
    cooldown_end_time: str
    expires_at: str
    goal: int
    id: str
    last_contribution: HypeTrainEventsResponseContributions
    level: int
    started_at: str
    top_contributions: list[HypeTrainEventsResponseContributions]
    total: int


class HypeTrainEventsResponseData(TypedDict):
    id: str
    event_type: str
    event_timestamp: str
    version: str
    event_data: HypeTrainEventsResponseEventData


class HypeTrainEventsResponse(TypedDict):
    data: list[HypeTrainEventsResponseData]
    pagination: Pagination


class CheckAutomodStatusResponseData(TypedDict):
    msg_id: str
    is_permitted: bool


class CheckAutomodStatusResponse(TypedDict):
    data: list[CheckAutomodStatusResponseData]


class AutomodSettingsResponseData(TypedDict):
    broadcaster_id: str
    moderator_id: str
    overall_level: int | None
    disability: int
    aggression: int
    sexuality_sex_or_gender: int
    misogyny: int
    bullying: int
    swearing: int
    race_ethnicity_or_religion: int
    sex_based_terms: int


class AutomodSettingsResponse(TypedDict):
    data: list[AutomodSettingsResponseData]


class UpdateAutomodSettingsResponseData(TypedDict):
    broadcaster_id: str
    moderator_id: str
    overall_level: int | None
    disability: int
    aggression: int
    sexuality_sex_or_gender: int
    misogyny: int
    bullying: int
    swearing: int
    race_ethnicity_or_religion: int
    sex_based_terms: int


class UpdateAutomodSettingsResponse(TypedDict):
    data: list[UpdateAutomodSettingsResponseData]


class BannedUsersResponseData(TypedDict):
    user_id: str
    user_login: str
    user_name: str
    expires_at: str
    created_at: str
    reason: str
    moderator_id: str
    moderator_login: str
    moderator_name: str


class BannedUsersResponse(TypedDict):
    data: list[BannedUsersResponseData]
    pagination: Pagination


class BanUserResponseData(TypedDict):
    broadcaster_id: str
    moderator_id: str
    user_id: str
    created_at: str
    end_time: str | None


class BanUserResponse(TypedDict):
    data: list[BanUserResponseData]


class UnbanRequestsResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    moderator_id: str
    moderator_login: str
    moderator_name: str
    user_id: str
    user_login: str
    user_name: str
    text: str
    status: Literal["pending", "approved", "denied", "acknowledged", "canceled"]
    created_at: str
    resolved_at: str
    resolution_text: str


class UnbanRequestsResponse(TypedDict):
    data: list[UnbanRequestsResponseData]
    pagination: Pagination


class ResolveUnbanRequestsResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    moderator_id: str
    moderator_login: str
    moderator_name: str
    user_id: str
    user_login: str
    user_name: str
    text: str
    status: Literal["approved", "denied"]
    created_at: str
    resolved_at: str
    resolution_text: str


class ResolveUnbanRequestsResponse(TypedDict):
    data: list[ResolveUnbanRequestsResponseData]


class BlockedTermsResponseData(TypedDict):
    broadcaster_id: str
    moderator_id: str
    id: str
    text: str
    created_at: str
    updated_at: str
    expires_at: str | None


class BlockedTermsResponse(TypedDict):
    data: list[BlockedTermsResponseData]
    pagination: Pagination


class AddBlockedTermResponse(TypedDict):
    data: list[BlockedTermsResponseData]


class ModeratedChannelsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str


class ModeratedChannelsResponse(TypedDict):
    data: list[ModeratedChannelsResponseData]
    pagination: Pagination


class ModeratorsResponseData(TypedDict):
    user_id: str
    user_login: str
    user_name: str


class ModeratorsResponse(TypedDict):
    data: list[ModeratorsResponseData]
    pagination: Pagination


class VipsResponseData(TypedDict):
    user_id: str
    user_name: str
    user_login: str


class VipsResponse(TypedDict):
    data: list[VipsResponseData]
    pagination: Pagination


class ShieldModeStatusResponseData(TypedDict):
    is_active: bool
    moderator_id: str
    moderator_login: str
    moderator_name: str
    last_activated_at: str


class UpdateShieldModeStatusResponse(TypedDict):
    data: list[ShieldModeStatusResponseData]


class ShieldModeStatusResponse(TypedDict):
    data: list[ShieldModeStatusResponseData]


class PollsResponseChoices(TypedDict):
    id: str
    title: str
    votes: int
    channel_points_votes: int
    bits_votes: int


class PollsResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str
    title: str
    choices: list[PollsResponseChoices]
    bits_voting_enabled: bool
    bits_per_vote: int
    channel_points_voting_enabled: bool
    channel_points_per_vote: int
    status: Literal["ACTIVE", "COMPLETED", "TERMINATED", "ARCHIVED", "MODERATED", "INVALID"]
    duration: int
    started_at: str
    ended_at: str | None


class PollsResponse(TypedDict):
    data: list[PollsResponseData]
    pagination: Pagination


class CreatePollResponse(TypedDict):
    data: list[PollsResponseData]


class EndPollResponse(TypedDict):
    data: list[PollsResponseData]


class PredictionsResponseTopPredictors(TypedDict):
    user_id: str
    user_name: str
    user_login: str
    channel_points_used: int
    channel_points_won: int


class PredictionsResponseOutcomes(TypedDict):
    id: str
    title: str
    users: int
    channel_points: int
    top_predictors: list[PredictionsResponseTopPredictors]
    color: Literal["BLUE", "PINK"]


class PredictionsResponseData(TypedDict):
    id: str
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str
    title: str
    winning_outcome_id: str | None
    outcomes: list[PredictionsResponseOutcomes]
    prediction_window: int
    status: Literal["ACTIVE", "CANCELED", "LOCKED", "RESOLVED"]
    created_at: str
    ended_at: str | None
    locked_at: str | None


class PredictionsResponse(TypedDict):
    data: list[PredictionsResponseData]
    pagination: Pagination


class CreatePredictionResponse(TypedDict):
    data: list[PredictionsResponseData]


class EndPredictionResponse(TypedDict):
    data: list[PredictionsResponseData]


class StartARaidResponseData(TypedDict):
    created_at: str
    is_mature: bool


class StartARaidResponse(TypedDict):
    data: list[StartARaidResponseData]


class ChannelStreamScheduleResponseCategory(TypedDict):
    id: str
    name: str


class ChannelStreamScheduleResponseSegments(TypedDict):
    id: str
    start_time: str
    end_time: str
    title: str
    canceled_until: str | None
    category: ChannelStreamScheduleResponseCategory | None
    is_recurring: bool


class ChannelStreamScheduleResponseVacation(TypedDict):
    start_time: str
    end_time: str


class ChannelStreamScheduleResponseData(TypedDict):
    segments: list[ChannelStreamScheduleResponseSegments]
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str
    vacation: ChannelStreamScheduleResponseVacation | None


class ChannelStreamScheduleResponse(TypedDict):
    data: ChannelStreamScheduleResponseData
    pagination: Pagination


class CreateChannelStreamScheduleSegmentResponse(TypedDict):
    data: ChannelStreamScheduleResponseData


class UpdateChannelStreamScheduleSegmentResponse(TypedDict):
    data: ChannelStreamScheduleResponseData


class SearchCategoriesResponseData(TypedDict):
    box_art_url: str
    name: str
    id: str


class SearchCategoriesResponse(TypedDict):
    data: list[SearchCategoriesResponseData]


class SearchChannelsResponseData(TypedDict):
    broadcaster_language: str
    broadcaster_login: str
    display_name: str
    game_id: str
    game_name: str
    id: str
    is_live: bool
    tag_ids: list[str]
    tags: list[str]
    thumbnail_url: str
    title: str
    started_at: str


class SearchChannelsResponse(TypedDict):
    data: list[SearchChannelsResponseData]
    pagination: Pagination


class StreamKeyResponseData(TypedDict):
    stream_key: str


class StreamKeyResponse(TypedDict):
    data: list[StreamKeyResponseData]


class StreamsResponseData(TypedDict):
    id: str
    user_id: str
    user_login: str
    user_name: str
    # Game ID and name are empty when no category is set
    game_id: str
    game_name: str
    type: Literal["live", ""]
    title: str
    tags: list[str]
    viewer_count: int
    started_at: str
    language: str
    thumbnail_url: str
    tag_ids: list[str]
    is_mature: bool


class StreamsResponse(TypedDict):
    data: list[StreamsResponseData]
    pagination: Pagination


class FollowedStreamsResponse(TypedDict):
    data: list[StreamsResponseData]
    pagination: Pagination


class CreateStreamMarkerResponseData(TypedDict):
    id: str
    created_at: str
    position_seconds: int
    description: str


class CreateStreamMarkerResponse(TypedDict):
    data: list[CreateStreamMarkerResponseData]


class StreamMarkersResponseMarkers(TypedDict):
    id: str
    created_at: str
    description: str
    position_seconds: int
    url: str


class StreamMarkersResponseVideos(TypedDict):
    video_id: str
    markers: list[StreamMarkersResponseMarkers]


class StreamMarkersResponseData(TypedDict):
    user_id: str
    user_name: str
    user_login: str
    videos: list[StreamMarkersResponseVideos]


class StreamMarkersResponse(TypedDict):
    data: list[StreamMarkersResponseData]
    pagination: Pagination


class BroadcasterSubscriptionsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    gifter_id: str
    gifter_login: str
    gifter_name: str
    is_gift: bool
    plan_name: str
    tier: Literal["1000", "2000", "3000"]
    user_id: str
    user_name: str
    user_login: str


class BroadcasterSubscriptionsResponse(TypedDict):
    data: list[BroadcasterSubscriptionsResponseData]
    pagination: Pagination
    points: int
    total: int


class CheckUserSubscriptionResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    gifter_id: str | None
    gifter_login: str | None
    gifter_name: str | None
    is_gift: bool
    tier: Literal["1000", "2000", "3000"]


class CheckUserSubscriptionResponse(TypedDict):
    data: list[CheckUserSubscriptionResponseData]


class ChannelTeamsResponseData(TypedDict):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    background_image_url: str
    banner: str
    created_at: str
    updated_at: str
    info: str
    thumbnail_url: str
    team_name: str
    team_display_name: str
    id: str


class ChannelTeamsResponse(TypedDict):
    data: list[ChannelTeamsResponseData]


class TeamsResponseUsers(TypedDict):
    user_id: str
    user_login: str
    user_name: str


class TeamsResponseData(TypedDict):
    users: list[TeamsResponseUsers]
    background_image_url: str
    banner: str
    created_at: str
    updated_at: str
    info: str
    thumbnail_url: str
    team_name: str
    team_display_name: str
    id: str


class TeamsResponse(TypedDict):
    data: list[TeamsResponseData]


class UsersResponseData(TypedDict):
    id: str
    login: str
    display_name: str
    type: Literal["admin", "global_mod", "staff", ""]
    broadcaster_type: Literal["affiliate", "partner", ""]
    description: str
    profile_image_url: str
    offline_image_url: str
    view_count: int
    email: str
    created_at: str


class UsersResponse(TypedDict):
    data: list[UsersResponseData]


class UpdateUserResponse(TypedDict):
    data: list[UsersResponseData]


class UserBlockListResponseData(TypedDict):
    user_id: str
    user_login: str
    display_name: str


class UserBlockListResponse(TypedDict):
    data: list[UserBlockListResponseData]


class UserExtensionsResponseData(TypedDict):
    id: str
    version: str
    name: str
    can_activate: bool
    type: list[Literal["component", "mobile", "overlay", "panel"]]


class UserExtensionsResponse(TypedDict):
    data: list[UserExtensionsResponseData]


class UserPanelItem(TypedDict):
    active: bool
    id: str
    version: str
    name: str


class UserPanelOverlayItem(TypedDict):
    active: bool
    id: str
    version: str
    name: str


class UserPanelComponentItem(TypedDict):
    active: bool
    id: str
    version: str
    name: str
    x: int
    y: int


class UserActiveExtensionsResponseData(TypedDict):
    panel: dict[str, UserPanelItem]
    overlay: dict[str, UserPanelOverlayItem]
    component: dict[str, UserPanelComponentItem]


class UserActiveExtensionsResponse(TypedDict):
    data: UserActiveExtensionsResponseData


class UpdateUserExtensionsResponse(TypedDict):
    data: UserActiveExtensionsResponseData


class VideosResponseMutedSegments(TypedDict):
    duration: int
    offset: int


class VideosResponseData(TypedDict):
    id: str
    stream_id: str | None
    user_id: str
    user_login: str
    user_name: str
    title: str
    description: str
    created_at: str
    published_at: str
    url: str
    thumbnail_url: str
    viewable: str
    view_count: int
    language: str
    type: Literal["archive", "highlight", "upload"]
    duration: str
    muted_segments: list[VideosResponseMutedSegments] | None


class VideosResponse(TypedDict):
    data: list[VideosResponseData]
    pagination: Pagination


class DeleteVideosResponse(TypedDict):
    data: list[str]


class WarnChatUserResponseData(TypedDict):
    broadcaster_id: str
    user_id: str
    moderator_id: str
    reason: str


class WarnChatUserResponse(TypedDict):
    data: list[WarnChatUserResponseData]



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import CreatorGoalsResponseData

__all__ = ("Goal",)


class Goal:
    """Represents a broadcaster's goal.

    The goal's ``type`` determines how the ``current_target`` is increased or decreased.

    - If type is ``follower``, this field is set to the broadcaster's current number of followers. This number increases with new followers and decreases when users unfollow the broadcaster.

    - If type is ``subscription``, this field is increased and decreased by the points value associated with the subscription tier. For example, if a tier-two subscription is worth 2 points, this field is increased or decreased by 2, not 1.

    - If type is ``subscription_count``, this field is increased by 1 for each new subscription and decreased by 1 for each user that unsubscribes.

    - If type is ``new_subscription``, this field is increased by the points value associated with the subscription tier. For example, if a tier-two subscription is worth 2 points, this field is increased by 2, not 1.

    - If type is ``new_subscription_count``, this field is increased by 1 for each new subscription.

    +------------------------+-----------------------------------------------------------------------------------------------------------------------+
    | Type                   | Description                                                                                                           |
    +========================+=======================================================================================================================+
    | follower               | The goal is to increase followers.                                                                                    |
    +------------------------+-----------------------------------------------------------------------------------------------------------------------+
    | subscription           | The goal is to increase subscriptions. This type shows the net increase or decrease in tier points associated with    |
    |                        | the subscriptions.                                                                                                    |
    +------------------------+-----------------------------------------------------------------------------------------------------------------------+
    | subscription_count     | The goal is to increase subscriptions. This type shows the net increase or decrease in the number of subscriptions.   |
    +------------------------+-----------------------------------------------------------------------------------------------------------------------+
    | new_subscription       | The goal is to increase subscriptions. This type shows only the net increase in tier points associated with the       |
    |                        | subscriptions (it does not account for users that unsubscribed since the goal started).                               |
    +------------------------+-----------------------------------------------------------------------------------------------------------------------+
    | new_subscription_count | The goal is to increase subscriptions. This type shows only the net increase in the number of subscriptions (it does  |
    |                        | not account for users that unsubscribed since the goal started).                                                      |
    +------------------------+-----------------------------------------------------------------------------------------------------------------------+

    Attributes
    -----------
    id: str
        An ID that identifies this goal.
    broadcaster: PartialUser
        The broadcaster that created the goal.
    type: typing.Literal["follower", "subscription", "subscription_count", "new_subscription", "new_subscription_count"]
        The type of goal. Possible values are: `follower`, `subscription`, `subscription_count`, `new_subscription`, `new_subscription_count`
        Please refer to the documentation for more details.
    description: str
        A description of the goal. Is an empty string if not specified.
    current_amount: int
        The goal's current value.
    target_amount: int
        The goal's target value. For example, if the broadcaster has 200 followers before creating the goal, and their goal is to double that number, this field is set to 400.
    created_at: datetime.datetime
        The datetime that the broadcaster created the goal.

    """

    __slots__ = ("broadcaster", "created_at", "current_amount", "description", "id", "target_amount", "type")

    def __init__(self, data: CreatorGoalsResponseData, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.type: Literal[
            "follower", "subscription", "subscription_count", "new_subscription", "new_subscription_count"
        ] = data["type"]
        self.description: str = data["description"]
        self.current_amount: int = data["current_amount"]
        self.target_amount: int = data["target_amount"]
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])

    def __repr__(self) -> str:
        return f"<Goal id={self.id} type={self.type} description={self.description}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .eventsub.enums import TransportMethod


if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

    from .authentication import Scopes
    from .eventsub.enums import SubscriptionType
    from .types_.eventsub import Condition, _SubscriptionData
    from .types_.tokens import _TokenRefreshedPayload


__all__ = ("EventErrorPayload", "TokenRefreshedPayload", "WebsocketSubscriptionData")


class EventErrorPayload:
    """Payload received in the :class:`Client.event_error` event when an error
    occurs during an event listener.

    Attributes
    ----------
    error: Exception
        The error that occurred.
    listener: Callable[..., Coroutine[Any, Any, None]]
        The listener that raised the error.
    original: Any
        The original event payload that was passed to the listener that caused the error.
    """

    __slots__ = ("error", "listener", "original")

    def __init__(self, *, error: Exception, listener: Callable[..., Coroutine[Any, Any, None]], original: Any) -> None:
        self.error: Exception = error
        self.listener: Callable[..., Coroutine[Any, Any, None]] = listener
        self.original: Any = original


class WebsocketSubscriptionData:
    """
    Payload returned from :meth:`~twitchio.Client.websocket_subscriptions` and
    :meth:`~twitchio.Client.delete_websocket_subscription` containing relevant information about the websocket subscription.

    Attributes
    ----------
    raw_data: dict[str, Any]
        The data contained in this class as a :class:`dict`
    condition: dict[str, str]
        The condition used to subscribe to the subscription as a :class:`dict`.
    token_for: str
        The User ID used to subscribe to this subscription. This could be the Client Users ID.
    transport: :class:`~twitchio.eventsub.TransportMethod`
        The :class:`~twitchio.eventsub.TransportMethod` enum.
        This will always be :attr:`~twitchio.eventsub.TransportMethod.WEBSOCKET`.
    type: :class:`~twitchio.eventsub.SubscriptionType`
        The subscription type as an enum value.
    version: str
        The version of the subscription on twitch.
    """

    __slots__ = (
        "condition",
        "raw_data",
        "token_for",
        "transport",
        "type",
        "version",
    )

    def __init__(self, data: _SubscriptionData) -> None:
        self.raw_data = data
        self.condition: Condition = data["condition"]
        self.token_for: str = data["token_for"]
        self.transport: TransportMethod = TransportMethod.WEBSOCKET
        self.type: SubscriptionType = data["type"]
        self.version: str = data["version"]


class TokenRefreshedPayload:
    """Payload received in the :func:`~twitchio.event_token_refreshed` event when a token managed by TwitchIO is successfully
    refreshed on the :class:`~twitchio.Client`.

    Attributes
    ----------
    user_id: str
        The User ID associated with the refreshed token.
    refresh_token: str
        The new refresh token returned by Twitch.
    token: str
        The updated token after refreshing returned by Twitch.
    scopes: :class:`~twitchio.Scopes`
        The scopes associated with the token.
    expires_in: int
        The time the new token expires as an :class:`int` of seconds.
    """

    __slots__ = (
        "expires_in",
        "refresh_token",
        "scopes",
        "token",
        "user_id",
    )

    def __init__(self, data: _TokenRefreshedPayload) -> None:
        self.user_id: str = data["user_id"]
        self.refresh_token: str = data["refresh_token"]
        self.token: str = data["token"]
        self.scopes: Scopes = data["scopes"]
        self.expires_in: int = data["expires_in"]



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from functools import partial
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, Self, TypeAlias, Unpack

from .core import Command, CommandErrorPayload


if TYPE_CHECKING:
    from collections.abc import Callable

    from .context import Context
    from .types_ import ComponentOptions


__all__ = ("Component",)


CoroC: TypeAlias = Coroutine[Any, Any, bool]


class _MetaComponent:
    __component_name__: str
    __component_extras__: dict[Any, Any]
    __component_specials__: ClassVar[list[str]] = []
    __all_commands__: dict[str, Command[Any, ...]]
    __all_listeners__: dict[str, list[Callable[..., Coroutine[Any, Any, None]]]]
    __all_guards__: list[Callable[..., bool] | Callable[..., CoroC]]

    @classmethod
    def _component_special(cls, obj: Any) -> Any:
        setattr(obj, "__component_special__", True)
        cls.__component_specials__.append(obj.__name__)

        return obj

    def __init_subclass__(cls, **kwargs: Unpack[ComponentOptions]) -> None:
        name: str | None = kwargs.get("name")
        if name:
            cls.__component_name__ = name

        cls.__component_extras__ = kwargs.get("extras", {})

    def __new__(cls, *args: Any, **Kwargs: Any) -> Self:
        self: Self = super().__new__(cls)

        if not hasattr(self, "__component_name__"):
            self.__component_name__ = cls.__qualname__

        commands: dict[str, Command[Any, ...]] = {}
        listeners: dict[str, list[Callable[..., Coroutine[Any, Any, None]]]] = {}
        guards: list[Callable[..., bool] | Callable[..., CoroC]] = []

        no_special: str = 'Commands, listeners and guards must not start with special name "component_" in components:'
        no_over: str = 'The special method "{}" can not be overriden in components.'

        for base in reversed(cls.__mro__):
            for name, member in base.__dict__.items():
                if name in self.__component_specials__ and not hasattr(member, "__component_special__"):
                    raise TypeError(no_over.format(name))

                if isinstance(member, Command):
                    if name.startswith("component_"):
                        raise TypeError(f'{no_special} "{member._callback.__qualname__}" is invalid.')  # type: ignore

                    if not member.extras:
                        member._extras = self.__component_extras__

                    if not member.parent:  # type: ignore
                        commands[name] = member

                elif hasattr(member, "__listener_name__"):
                    if name.startswith("component_"):
                        raise TypeError(f'{no_special} "{member.__qualname__}" is invalid.')

                    # Need to inject the component into the listener...
                    injected = partial(member, self)

                    try:
                        listeners[member.__listener_name__].append(injected)
                    except KeyError:
                        listeners[member.__listener_name__] = [injected]

                elif hasattr(member, "__component_guard__"):
                    if not member.__component_guard__:
                        continue

                    if name.startswith("component_"):
                        raise TypeError(f'{no_special} "{member.__qualname__}" is invalid.')

                    guards.append(member)

        cls.__all_commands__ = commands
        cls.__all_listeners__ = listeners
        cls.__all_guards__ = guards

        return self


class Component(_MetaComponent):
    """TwitchIO Component class.

    Components are a powerful class used to help organize and manage commands, events, guards and errors.

    This class inherits from a special metaclass which implements logic to help manage other parts of the TwitchIO
    commands extension together.

    The Component must be added to your bot via :meth:`.commands.Bot.add_component`. After this class has been added, all
    commands and event listeners contained within the component will also be added to your bot.

    You can remove this Component and all the commands and event listeners associated with it via
    :meth:`.commands.Bot.remove_component`.

    There are two built-in methods of components that aid in doing any setup or teardown, when they are added and removed
    respectfully.

    - :meth:`~.component_load`

    - :meth:`~.component_teardown`


    Below are some special methods of Components which are designed to be overriden when needed:

    - :meth:`~.component_load`

    - :meth:`~.component_teardown`

    - :meth:`~.component_command_error`

    - :meth:`~.component_before_invoke`

    - :meth:`~.component_after_invoke`


    Components also implement some special decorators which can only be used inside of Components. The decorators are
    class method decorators and won't need an instance of a Component to use.

    - :meth:`~.listener`

    - :meth:`~.guard`


    Commands can beed added to Components with their respected decorators, and don't need to be added to the bot, as they
    will be added when you add the component.

    - ``@commands.command()``

    - ``@commands.group()``


    .. note::

        This version of TwitchIO has not yet implemented the ``modules`` implementation of ``commands.ext``.
        This part of ``commands.ext`` will allow you to easily load and unload separate python files that could contain
        components.

    .. important::

        Due to the implementation of Components, you shouldn't make a call to ``super().__init__()`` if you implement an
        ``__init__`` on this component.

    Examples
    --------

    .. code:: python3

        class Bot(commands.Bot):
            # Do your required __init__ etc first...

            # You can use setup_hook to add components...
            async def setup_hook(self) -> None:
                await self.add_component(MyComponent())


        class MyComponent(commands.Component):

            # Some simple commands...
            @commands.command()
            async def hi(self, ctx: commands.Command) -> None:
                await ctx.send(f"Hello {ctx.chatter.mention}!")

            @commands.command()
            async def apple(self, ctx: commands.Command, *, count: int) -> None:
                await ctx.send(f"You have {count} apples?!")

            # An example of using an event listener in a component...
            @commands.Component.listener()
            async def event_message(self, message: twitchio.ChatMessage) -> None:
                print(f"Received Message in component: {message.content}")

            # An example of a before invoke hook that is executed directly before any command in this component...
            async def component_before_invoke(self, ctx: commands.Command) -> None:
                print(f"Processing command in component '{self.name}'.")
    """

    @property
    def name(self) -> str:
        """Property returning the name of this component, this is either the qualified name of the class or
        the custom provided name if set.
        """
        return self.__component_name__

    async def component_command_error(self, payload: CommandErrorPayload) -> bool | None:
        """Event called when an error occurs in a command in this Component.

        Similar to :meth:`~.commands.Bot.event_command_error` except only catches errors from commands within this Component.

        This method is intended to be overwritten, by default it does nothing.

        .. note::

            Explicitly returning ``False`` in this function will stop it being dispatched to any other error handler.
        """

    async def component_load(self) -> None:
        """Hook called when the component is about to be loaded into the bot.

        You should use this hook to do any async setup required when loading a component.
        See: :meth:`.component_teardown` for a hook called when a Component is unloaded from the bot.

        This method is intended to be overwritten, by default it does nothing.

        .. important::

            If this method raises or fails, the Component will **NOT** be loaded. Instead it will be cleaned up and removed
            and the error will propagate.
        """

    async def component_teardown(self) -> None:
        """Hook called when the component is about to be unloaded from the bot.

        You should use this hook to do any async teardown/cleanup required on the component.
        See: :meth:`.component_load` for a hook called when a Component is loaded into the bot.

        This method is intended to be overwritten, by default it does nothing.
        """

    async def component_before_invoke(self, ctx: Context) -> None:
        """Hook called before a :class:`~.commands.Command` in this Component is invoked.

        Similar to :meth:`~.commands.Bot.before_invoke` but only applies to commands in this Component.
        """

    async def component_after_invoke(self, ctx: Context) -> None:
        """Hook called after a :class:`~.commands.Command` has successfully invoked in this Component.

        Similar to :meth:`~.commands.Bot.after_invoke` but only applies to commands in this Component.
        """

    @_MetaComponent._component_special
    def extras(self) -> MappingProxyType[Any, Any]:
        """Property returning a :class:`types.MappingProxyType` of the extras applied to every command in this Component.

        See: :attr:`~.commands.Command.extras` for more information on :class:`~.commands.Command` extras.
        """
        return MappingProxyType(self.__component_extras__)

    @_MetaComponent._component_special
    def guards(self) -> list[Callable[..., bool] | Callable[..., CoroC]]:
        """Property returning the guards applied to every command in this Component.

        See: :func:`.commands.guard` for more information on guards and how to use them.

        See: :meth:`.guard` for a way to apply guards to every command in this Component.
        """
        return self.__all_guards__

    @classmethod
    def listener(cls, name: str | None = None) -> Any:
        """|deco|

        A decorator which adds an event listener similar to :meth:`~.commands.Bot.listener` but contained within this
        component.

        Event listeners in components can listen to any dispatched event, and don't interfere with their base implementation.
        See: :meth:`~.commands.Bot.listener` for more information on event listeners.

        By default, listeners use the name of the function wrapped for the event name. This can be changed by passing the
        name parameter.

        .. note::

            You can have multiple of the same event listener per component, see below for an example.

        Examples
        --------

            .. code:: python3

                # By default if no name parameter is passed, the name of the event listened to is the same as the function...

                class MyComponent(commands.Component):

                    @commands.Component.listener()
                    async def event_message(self, payload: twitchio.ChatMessage) -> None:
                        ...

            .. code:: python3

                # You can listen to two or more of the same event in a single component...
                # The name parameter should have the "event_" prefix removed...

                class MyComponent(commands.Component):

                    @commands.Component.listener("message")
                    async def event_message_one(self, payload: twitchio.ChatMessage) -> None:
                        ...

                    @commands.Component.listener("message")
                    async def event_message_two(self, payload: twitchio.ChatMessage) -> None:
                        ...

        Parameters
        ----------
        name: str
            The name of the event to listen to, E.g. ``"event_message"`` or simply ``"message"``.
        """

        def wrapper(func: Callable[..., Coroutine[Any, Any, None]]) -> Callable[..., Coroutine[Any, Any, None]]:
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f'Component listener func "{func.__qualname__}" must be a coroutine function.')

            name_ = name or func.__name__
            qual = f"event_{name_.removeprefix('event_')}"

            setattr(func, "__listener_name__", qual)
            return func

        return wrapper

    @classmethod
    def guard(cls) -> Any:
        """|deco|

        A decorator which wraps a standard function *or* coroutine function which should
        return either ``True`` or ``False``, and applies a guard to every :class:`~.commands.Command` in this component.

        The wrapped function should take in one parameter :class:`~.commands.Context` the context surrounding
        command invocation, and return a bool indicating whether a command should be allowed to run.

        If the wrapped function returns ``False``, the chatter will not be able to invoke the command and an error will be
        raised. If the wrapped function returns ``True`` the chatter will be able to invoke the command,
        assuming all the other guards also pass their predicate checks.

        See: :func:`~.commands.guard` for more information on guards, what they do and how to use them.

        See: :meth:`~.commands.Bot.global_guard` for a global guard, applied to every command the bot has added.

        Example
        -------

        .. code:: python3

            class NotModeratorError(commands.GuardFailure):
                ...

            class MyComponent(commands.Component):

                # The guard below will be applied to every command contained in your component...
                # This guard raises our custom exception for easily identifying the error in our handler...

                @commands.Component.guard()
                def is_moderator(self, ctx: commands.Context) -> bool:
                    if not ctx.chatter.moderator:
                        raise NotModeratorError

                    return True

                @commands.command()
                async def test(self, ctx: commands.Context) -> None:
                    await ctx.reply(f"You are a moderator of {ctx.channel}")

                async def component_command_error(self, payload: commands.CommandErrorPayload) -> bool | None:
                    error = payload.exception
                    ctx = payload.context

                    if isinstance(error, NotModeratorError):
                        await ctx.reply("Only moderators can use this command!")

                        # This explicit False return stops the error from being dispatched anywhere else...
                        return False

        Raises
        ------
        GuardFailure
            The guard predicate returned ``False`` and prevented the chatter from using the command.
        """

        def wrapper(func: Callable[..., bool] | Callable[..., CoroC]) -> Callable[..., bool] | Callable[..., CoroC]:
            setattr(func, "__component_guard__", True)
            return func

        return wrapper
